import unittest

from kinto.core.errors import ERRORS
from kinto.core.testing import FormattedErrorMixin

from .support import MINIMALIST_BUCKET, MINIMALIST_GROUP, BaseWebTest


class GroupViewTest(FormattedErrorMixin, BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/groups"
    record_url = "/buckets/beers/groups/moderators"

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        resp = self.app.put_json(self.record_url, MINIMALIST_GROUP, headers=self.headers)
        self.record = resp.json["data"]

    def test_collection_endpoint_lists_them_all(self):
        resp = self.app.get(self.collection_url, headers=self.headers)
        records = resp.json["data"]
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]["members"], ["fxa:user"])

    def test_groups_can_be_posted_without_id(self):
        resp = self.app.post_json(
            self.collection_url, MINIMALIST_GROUP, headers=self.headers, status=201
        )
        self.assertIn("id", resp.json["data"])
        self.assertEqual(resp.json["data"]["members"], ["fxa:user"])

    def test_groups_can_be_put_with_empty_body(self):
        self.app.put("/buckets/beers/groups/simple", headers=self.headers)

    def test_groups_can_be_created_with_just_permissions(self):
        group = {"permissions": {"write": ["github:me"]}}
        resp = self.app.put_json("/buckets/beers/groups/g", group, headers=self.headers)
        self.assertNotIn("members", resp.json["data"])

    def test_groups_can_be_create_without_members_attribute(self):
        group = {"data": {"alias": "admins"}}
        resp = self.app.put_json(self.record_url, group, headers=self.headers)
        self.assertEqual(resp.json["data"]["members"], [])

    def test_groups_can_be_patched_without_members_attribute(self):
        group = {"data": {"alias": "admins"}}
        resp = self.app.patch_json(self.record_url, group, headers=self.headers)
        self.assertEqual(resp.json["data"]["members"], ["fxa:user"])
        self.assertEqual(resp.json["data"]["alias"], "admins")

    def test_groups_can_be_put_with_simple_name(self):
        self.assertEqual(self.record["id"], "moderators")

    def test_groups_name_should_be_simple(self):
        self.app.put_json(
            "/buckets/beers/groups/__moderator__",
            MINIMALIST_GROUP,
            headers=self.headers,
            status=400,
        )

    def test_groups_can_have_arbitrary_attributes(self):
        mailinglist = "kinto@mozilla.com"
        group = {
            **MINIMALIST_GROUP,
            "data": {**MINIMALIST_GROUP["data"], "mailinglist": mailinglist},
        }
        resp = self.app.put_json("/buckets/beers/groups/moderator", group, headers=self.headers)
        data = resp.json["data"]
        self.assertIn("mailinglist", data)
        self.assertEqual(data["mailinglist"], mailinglist)

    def test_groups_can_be_filtered_by_arbitrary_attribute(self):
        group = {**MINIMALIST_GROUP, "data": {**MINIMALIST_GROUP["data"], "size": 3}}
        self.app.put_json("/buckets/beers/groups/moderator", group, headers=self.headers)
        resp = self.app.get("/buckets/beers/groups?has_size=true&min_size=2", headers=self.headers)
        data = resp.json["data"]
        self.assertEqual(len(data), 1)

    def test_groups_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, "Content-Type": "text/plain"}
        self.app.put(
            "/buckets/beers/groups/moderator", MINIMALIST_GROUP, headers=headers, status=415
        )

    def test_unknown_bucket_raises_403(self):
        other_bucket = self.collection_url.replace("beers", "sodas")
        self.app.get(other_bucket, headers=self.headers, status=403)

    def test_groups_are_isolated_by_bucket(self):
        other_bucket = self.record_url.replace("beers", "sodas")
        self.app.put_json("/buckets/sodas", MINIMALIST_BUCKET, headers=self.headers)
        self.app.get(other_bucket, headers=self.headers, status=404)

    def test_wrong_create_permissions_cannot_be_added_on_groups(self):
        group = {**MINIMALIST_GROUP, "permissions": {"group:create": ["fxa:user"]}}
        self.app.put_json(
            "/buckets/beers/groups/moderator", group, headers=self.headers, status=400
        )

    def test_recreate_group_after_deletion_returns_a_201(self):
        self.app.put_json(
            "/buckets/beers/groups/moderator", MINIMALIST_GROUP, headers=self.headers, status=201
        )
        self.app.delete("/buckets/beers/groups/moderator", headers=self.headers, status=200)
        self.app.put_json(
            "/buckets/beers/groups/moderator", MINIMALIST_GROUP, headers=self.headers, status=201
        )

    def test_group_doesnt_accept_system_Everyone(self):
        group = {**MINIMALIST_GROUP, "data": {"members": ["system.Everyone"]}}
        response = self.app.put_json(
            "/buckets/beers/groups/moderator", group, headers=self.headers, status=400
        )
        self.assertFormattedError(
            response,
            400,
            ERRORS.INVALID_PARAMETERS,
            "Invalid parameters",
            "'system.Everyone' is not a valid user ID.",
        )

    def test_group_doesnt_accept_groups_inside_groups(self):
        group = {**MINIMALIST_GROUP, "data": {"members": ["/buckets/beers/groups/administrators"]}}
        response = self.app.put_json(
            "/buckets/beers/groups/moderator", group, headers=self.headers, status=400
        )
        self.assertFormattedError(
            response,
            400,
            ERRORS.INVALID_PARAMETERS,
            "Invalid parameters",
            "'/buckets/beers/groups/administrators' is not a valid user ID.",
        )


class GroupManagementTest(BaseWebTest, unittest.TestCase):
    group_url = "/buckets/beers/groups/moderators"

    def setUp(self):
        super().setUp()
        self.create_bucket("beers")

    def test_groups_can_be_deleted(self):
        self.create_group("beers", "moderators")
        self.app.delete(self.group_url, headers=self.headers)
        self.app.get(self.group_url, headers=self.headers, status=404)

    def test_unknown_group_raises_404(self):
        other_group = self.group_url.replace("moderators", "blah")
        resp = self.app.get(other_group, headers=self.headers, status=404)
        self.assertEqual(resp.json["details"]["id"], "blah")
        self.assertEqual(resp.json["details"]["resource_name"], "group")

    def test_group_is_removed_from_users_principals_on_group_deletion(self):
        self.app.put_json(self.group_url, MINIMALIST_GROUP, headers=self.headers, status=201)
        self.assertIn(self.group_url, self.permission.get_user_principals("fxa:user"))
        self.app.delete(self.group_url, headers=self.headers, status=200)
        self.assertNotIn(self.group_url, self.permission.get_user_principals("fxa:user"))

    def test_group_is_removed_from_users_principals_on_groups_deletion(self):
        self.create_group("beers", "moderators", ["natim", "fxa:me"])
        self.create_group("beers", "reviewers", ["natim", "alexis"])

        self.app.delete("/buckets/beers/groups", headers=self.headers, status=200)

        self.assertEqual(self.permission.get_user_principals("fxa:me"), set())
        self.assertEqual(self.permission.get_user_principals("natim"), set())
        self.assertEqual(self.permission.get_user_principals("alexis"), set())

    def test_group_is_added_to_user_principals_when_added_to_members(self):
        self.create_group("beers", "moderators", ["natim", "mat"])

        self.app.get("/buckets/beers/groups", headers=self.headers, status=200)
        self.assertEqual(
            self.permission.get_user_principals("natim"), {"/buckets/beers/groups/moderators"}
        )
        self.assertEqual(
            self.permission.get_user_principals("mat"), {"/buckets/beers/groups/moderators"}
        )

    def test_group_is_added_to_user_principals_on_members_add_with_patch(self):
        self.create_group("beers", "moderators", ["natim", "mat"])
        group_url = "/buckets/beers/groups/moderators"
        group = {"data": {"members": ["natim", "mat", "alice"]}}
        self.app.patch_json(group_url, group, headers=self.headers, status=200)
        self.app.get("/buckets/beers/groups", headers=self.headers, status=200)
        self.assertEqual(self.permission.get_user_principals("natim"), {group_url})
        self.assertEqual(self.permission.get_user_principals("mat"), {group_url})
        self.assertEqual(self.permission.get_user_principals("alice"), {group_url})

    def test_group_member_removal_updates_user_principals(self):
        self.create_group("beers", "moderators", ["natim", "mat"])
        group_url = "/buckets/beers/groups/moderators"
        group = {"data": {"members": ["mat"]}}
        self.app.put_json(group_url, group, headers=self.headers, status=200)
        self.app.get("/buckets/beers/groups", headers=self.headers, status=200)
        self.assertEqual(self.permission.get_user_principals("natim"), set())
        self.assertEqual(self.permission.get_user_principals("mat"), {group_url})

    def test_group_with_authenticated_is_added_to_everbody(self):
        self.create_group("beers", "reviewers", ["system.Authenticated"])
        self.assertEqual(
            self.permission.get_user_principals("natim"), {"/buckets/beers/groups/reviewers"}
        )
        self.assertEqual(
            self.permission.get_user_principals("mat"), {"/buckets/beers/groups/reviewers"}
        )

    def test_group_member_removal_updates_user_principals_with_patch(self):
        self.create_group("beers", "moderators", ["natim", "mat"])
        group_url = "/buckets/beers/groups/moderators"
        group = {"data": {"members": ["mat"]}}
        self.app.patch_json(group_url, group, headers=self.headers, status=200)
        self.assertEqual(self.permission.get_user_principals("natim"), set())
        self.assertEqual(self.permission.get_user_principals("mat"), {group_url})

    def test_groups_can_be_created_after_deletion(self):
        self.create_group("beers", "moderators")
        group_url = "/buckets/beers/groups/moderators"
        self.app.delete(group_url, headers=self.headers)
        headers = {**self.headers, "If-None-Match": "*"}
        self.app.put_json(group_url, MINIMALIST_GROUP, headers=headers, status=201)

    def test_groups_data_is_optional_with_patch(self):
        valid = {"permissions": {"write": ["github:me"]}}
        self.app.put_json(self.group_url, MINIMALIST_GROUP, headers=self.headers)
        self.app.patch_json(self.group_url, valid, headers=self.headers)



import unittest

from kinto.core.testing import skip_if_no_postgresql

from .support import PostgreSQLTest


@skip_if_no_postgresql
class PaginationTest(PostgreSQLTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["storage_max_fetch_size"] = 4
        return settings

    def setUp(self):
        super().setUp()
        for i in range(10):
            self.app.post_json("/mushrooms", {"data": {"name": str(i)}}, headers=self.headers)

    def test_storage_max_fetch_size_is_per_page(self):
        resp = self.app.get("/mushrooms?_limit=6", headers=self.headers)
        self.assertIn("Next-Page", resp.headers)
        self.assertEqual(len(resp.json["data"]), 4)

        next_page_url = resp.headers["Next-Page"].replace("http://localhost/v0", "")
        resp = self.app.get(next_page_url, headers=self.headers)
        self.assertIn("Next-Page", resp.headers)
        self.assertEqual(len(resp.json["data"]), 4)

        next_page_url = resp.headers["Next-Page"].replace("http://localhost/v0", "")
        resp = self.app.get(next_page_url, headers=self.headers)
        self.assertNotIn("Next-Page", resp.headers)
        self.assertEqual(len(resp.json["data"]), 2)



from bravado_core.response import validate_response

from .support import MINIMALIST_COLLECTION, OpenAPITest


class OpenAPICollectionResponsesTest(OpenAPITest):
    def test_get_collection_200(self):
        response = self.app.get("/buckets/b1/collections/c1", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].get_collection
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_post_collection_200(self):
        response = self.app.post_json(
            "/buckets/b1/collections", self.collection, headers=self.headers, status=200
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].create_collection
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_post_collection_201(self):
        response = self.app.post_json(
            "/buckets/b1/collections", MINIMALIST_COLLECTION, headers=self.headers, status=201
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].create_collection
        schema = self.spec.deref(op.op_spec["responses"]["201"])
        validate_response(schema, op, response)

    def test_put_collection_200(self):
        response = self.app.put_json(
            "/buckets/b1/collections/c1", MINIMALIST_COLLECTION, headers=self.headers, status=200
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].update_collection
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_put_collection_201(self):
        response = self.app.put_json(
            "/buckets/b1/collections/c2", MINIMALIST_COLLECTION, headers=self.headers, status=201
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].update_collection
        schema = self.spec.deref(op.op_spec["responses"]["201"])
        validate_response(schema, op, response)

    def test_delete_collection_200(self):
        response = self.app.delete("/buckets/b1/collections/c1", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].delete_collection
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_get_collections_200(self):
        response = self.app.get("/buckets/b1/collections", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].get_collections
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)

    def test_delete_collections_200(self):
        response = self.app.delete("/buckets/b1/collections", headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].delete_collections
        schema = self.spec.deref(op.op_spec["responses"]["200"])
        validate_response(schema, op, response)



"""
kinto.plugins.quotas.scripts: scripts to maintain quotas and fix them when they're broken
"""

import logging

from kinto.core.storage import Sort
from kinto.core.storage.utils import paginated

from .listener import BUCKET_QUOTA_OBJECT_ID, COLLECTION_QUOTA_OBJECT_ID
from .utils import record_size


logger = logging.getLogger(__name__)

OLDEST_FIRST = Sort("last_modified", 1)


def rebuild_quotas(storage, dry_run=False):
    for bucket in paginated(storage, resource_name="bucket", parent_id="", sorting=[OLDEST_FIRST]):
        bucket_id = bucket["id"]
        bucket_path = f"/buckets/{bucket['id']}"
        bucket_collection_count = 0
        bucket_record_count = 0
        bucket_storage_size = record_size(bucket)

        for collection in paginated(
            storage, resource_name="collection", parent_id=bucket_path, sorting=[OLDEST_FIRST]
        ):
            collection_info = rebuild_quotas_collection(storage, bucket_id, collection, dry_run)
            (collection_record_count, collection_storage_size) = collection_info
            bucket_collection_count += 1
            bucket_record_count += collection_record_count
            bucket_storage_size += collection_storage_size

        bucket_record = {
            "record_count": bucket_record_count,
            "storage_size": bucket_storage_size,
            "collection_count": bucket_collection_count,
        }
        if not dry_run:
            storage.update(
                resource_name="quota",
                parent_id=bucket_path,
                object_id=BUCKET_QUOTA_OBJECT_ID,
                obj=bucket_record,
            )

        logger.info(
            f"Bucket {bucket_id}. Final size: {bucket_collection_count} collections, {bucket_record_count} records, {bucket_storage_size} bytes."
        )


def rebuild_quotas_collection(storage, bucket_id, collection, dry_run=False):
    """Helper method for rebuild_quotas that updates a single collection."""
    collection_id = collection["id"]
    collection_record_count = 0
    collection_storage_size = record_size(collection)
    collection_uri = f"/buckets/{bucket_id}/collections/{collection_id}"
    for record in paginated(
        storage, resource_name="record", parent_id=collection_uri, sorting=[OLDEST_FIRST]
    ):
        collection_record_count += 1
        collection_storage_size += record_size(record)

    logger.info(
        f"Bucket {bucket_id}, collection {collection_id}. Final size: {collection_record_count} records, {collection_storage_size} bytes."
    )
    new_quota_info = {
        "record_count": collection_record_count,
        "storage_size": collection_storage_size,
    }
    if not dry_run:
        storage.update(
            resource_name="quota",
            parent_id=collection_uri,
            object_id=COLLECTION_QUOTA_OBJECT_ID,
            obj=new_quota_info,
        )
    return (collection_record_count, collection_storage_size)



import os

from kinto.core.decorators import cache_forever


HERE = os.path.dirname(__file__)


# Configured home page
@cache_forever
def admin_home_view(request):
    """
    This view reads the ``index.html`` file from the Admin assets path folder
    and serves it.

    This requires the Admin UI to be built with ``ASSET_PATH="/v1/admin/"``.
    """
    # Default location of the Admin UI is relative to this plugin source folder,
    # as pulled with the ``make pull-kinto-admin`` command.
    admin_assets_path = request.registry.settings["admin_assets_path"] or os.path.join(
        HERE, "build"
    )
    try:
        with open(os.path.join(admin_assets_path, "index.html")) as f:
            page_content = f.read()
    except FileNotFoundError:  # pragma: no cover
        with open(os.path.join(HERE, "public", "help.html")) as f:
            page_content = f.read()

    # Add Content-Security-Policy HTTP response header to protect against XSS:
    # only allow from local domain:
    allow_local_only = "; ".join(
        (
            "default-src 'self'",
            "img-src data: 'self'",
            "script-src 'self' 'unsafe-inline' 'unsafe-eval'",
            "style-src 'self' 'unsafe-inline'",
        )
    )
    request.response.headers["Content-Security-Policy"] = allow_local_only

    return page_content



from pyramid import httpexceptions

from kinto.core.errors import ERRORS

from . import BaseTest


class NotModifiedTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.stored = self.model.create_object({})

        self.resource = self.resource_class(request=self.get_request(), context=self.get_context())
        self.resource.request.validated = {**self.validated}
        self.resource.plural_get()
        self.validated = self.resource.request.validated
        current = self.last_response.headers["ETag"][1:-1]
        self.validated["header"]["If-None-Match"] = int(current)

    def test_plural_returns_200_if_change_meanwhile(self):
        self.validated["header"]["If-None-Match"] = 42
        self.resource.plural_get()  # Not raising.

    def test_plural_returns_304_if_no_change_meanwhile(self):
        try:
            self.resource.plural_get()
        except httpexceptions.HTTPNotModified as e:
            error = e
        self.assertEqual(error.code, 304)
        self.assertIsNotNone(error.headers.get("ETag"))
        self.assertIsNotNone(error.headers.get("Last-Modified"))

    def test_single_object_returns_304_if_no_change_meanwhile(self):
        self.resource.object_id = self.stored["id"]
        try:
            self.resource.get()
        except httpexceptions.HTTPNotModified as e:
            error = e
        self.assertEqual(error.code, 304)
        self.assertIsNotNone(error.headers.get("ETag"))
        self.assertIsNotNone(error.headers.get("Last-Modified"))

    def test_single_object_last_modified_is_returned(self):
        self.resource.timestamp = 0
        self.resource.object_id = self.stored["id"]
        try:
            self.resource.get()
        except httpexceptions.HTTPNotModified as e:
            error = e
        self.assertNotIn("1970", error.headers["Last-Modified"])


class ModifiedMeanwhileTest(BaseTest):
    def setUp(self):
        super().setUp()

        # Create an object using model (will have an incremented last_modified)
        self.stored = self.model.create_object({})
        self.stored.pop(self.model.permissions_field)

        # Update the object we just created (will set last_modified with the server ETag)
        self.resource.object_id = self.stored["id"]
        self.resource.put()
        # Update our object with last_modified provided by the server
        self.stored["last_modified"] = int(self.last_response.headers["ETag"][1:-1])

        self.validated = self.resource.request.validated
        self.current = int(self.last_response.headers["ETag"][1:-1])
        previous = self.current - 10
        self.validated["header"]["If-Match"] = previous

    def test_preconditions_errors_are_json_formatted(self):
        try:
            self.resource.plural_get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertEqual(
            error.json,
            {
                "errno": ERRORS.MODIFIED_MEANWHILE.value,
                "message": "Resource was modified meanwhile",
                "code": 412,
                "error": "Precondition Failed",
            },
        )

    def test_plural_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_get)

    def test_plural_returns_412_if_if_match_is_superior(self):
        self.validated["header"]["If-Match"] = self.current + 10
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_get)

    def test_plural_returns_200_if_if_match_is_equal(self):
        self.validated["header"]["If-Match"] = self.current
        self.resource.plural_get()

    def test_412_errors_on_plural_endpoint_do_not_provide_existing_object(self):
        try:
            self.resource.plural_get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertNotIn("existing", error.json.get("details", {}))

    def test_412_on_plural_endpoint_has_last_modified_timestamp(self):
        try:
            self.resource.plural_get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertIsNotNone(error.headers.get("ETag"))
        self.assertIsNotNone(error.headers.get("Last-Modified"))

    def test_412_errors_on_object_provide_existing_data(self):
        try:
            self.resource.put()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertDictEqual(error.json["details"]["existing"], self.stored)

    def test_single_object_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.get)

    def test_412_on_single_object_has_last_modified_timestamp(self):
        try:
            self.resource.get()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertIsNotNone(error.headers.get("ETag"))
        self.assertIsNotNone(error.headers.get("Last-Modified"))

    def test_create_returns_412_if_changed_meanwhile(self):
        self.validated["body"] = {"data": {"field": "new"}}
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_post)

    def test_put_returns_412_if_changed_meanwhile(self):
        self.model.delete_object(self.stored)
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_put_returns_412_if_changed_and_none_match_present(self):
        self.validated["body"] = {"data": {"field": "new"}}
        self.validated["header"]["If-None-Match"] = 42
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_put_returns_last_modified_if_changed_meanwhile(self):
        self.resource.timestamp = 0
        self.resource.object_id = self.stored["id"]
        try:
            self.resource.put()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertNotIn("1970", error.headers["Last-Modified"])

    def test_put_returns_412_if_deleted_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_put_if_none_match_star_fails_if_object_exists(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_get_if_none_match_star_fails_if_object_exists(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.get)

    def test_put_if_none_match_star_succeeds_if_object_does_not_exist(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.validated["body"] = {"data": {"field": "new"}}
        self.resource.object_id = self.resource.model.id_generator()
        self.resource.put()  # not raising.

    def test_put_if_none_match_star_succeeds_if_tombstone_exists(self):
        self.model.delete_object(self.stored)
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.validated["body"] = {"data": {"field": "new"}}
        self.resource.put()  # not raising.

    def test_post_if_none_match_star_fails_if_object_exists(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.resource.request.json = {"data": {"id": self.stored["id"], "field": "new"}}
        self.validated["body"] = self.resource.request.json
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_post)

    def test_post_if_none_match_star_succeeds_if_object_does_not_exist(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.resource.object_id = self.resource.model.id_generator()
        self.validated["body"] = {"data": {"id": self.resource.object_id, "field": "new"}}
        self.resource.plural_post()  # not raising.

    def test_get_if_none_match_star_fails_on_plural_endpoints(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_get)

    def test_delete_if_none_match_star_fails_on_plural_endpoints(self):
        self.validated["header"].pop("If-Match")
        self.validated["header"]["If-None-Match"] = "*"
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_delete)

    def test_get_if_match_star_succeeds_if_object_exists(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.object_id = self.stored["id"]
        self.resource.get()

    def test_post_if_match_star_succeeds_if_object_exists(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.request.json = {"data": {"id": self.stored["id"], "field": "new"}}
        self.validated["body"] = self.resource.request.json
        self.resource.plural_post()

    def test_put_if_match_star_succeeds_if_object_exists(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.put()

    def test_patch_if_match_star_succeeds_if_object_exists(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.request.json = {"data": {"id": self.stored["id"], "field": "new"}}
        self.validated["body"] = self.resource.request.json
        self.resource.patch()

    def test_delete_if_match_star_succeeds_if_object_exists(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.delete()

    def test_put_if_match_star_fails_if_object_does_not_exist(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.request.json = {"data": {"field": "new"}}
        self.validated["body"] = self.resource.request.json
        self.resource.object_id = self.resource.model.id_generator()
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_put_if_match_current_fails_if_object_does_not_exist(self):
        self.validated["header"]["If-Match"] = self.current  # wouldn't raise on existing
        self.resource.request.json = {"data": {"field": "new"}}
        self.validated["body"] = self.resource.request.json
        self.resource.object_id = self.resource.model.id_generator()
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_put_if_fails_if_if_match_is_superior(self):
        self.validated["header"]["If-Match"] += self.current + 10
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.put)

    def test_get_if_match_star_suceed_on_plural_endpoints(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.plural_get()

    def test_delete_if_match_star_suceed_on_plural_endpoints(self):
        self.validated["header"]["If-Match"] = "*"
        self.resource.plural_delete()

    def test_patch_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.patch)

    def test_patch_returns_last_modified_if_changed_meanwhile(self):
        self.resource.timestamp = 0
        try:
            self.resource.patch()
        except httpexceptions.HTTPPreconditionFailed as e:
            error = e
        self.assertNotIn("1970", error.headers["Last-Modified"])

    def test_delete_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.delete)

    def test_delete_all_returns_412_if_changed_meanwhile(self):
        self.assertRaises(httpexceptions.HTTPPreconditionFailed, self.resource.plural_delete)



from kinto.core.testing import unittest

from .support import BaseWebTest


class ContributeViewTest(BaseWebTest, unittest.TestCase):
    def test_returns_info_about_project(self):
        response = self.app.get("/contribute.json")
        keys = sorted(response.json.keys())
        expected = ["description", "keywords", "name", "participate", "repository", "urls"]
        self.assertEqual(keys, expected)



import unittest

from kinto.core.testing import get_user_headers

from .support import (
    MINIMALIST_BUCKET,
    MINIMALIST_COLLECTION,
    MINIMALIST_GROUP,
    MINIMALIST_RECORD,
    BaseWebTest,
)


RECORD_ID = "d5db6e57-2c10-43e2-96c8-56602ef01435"


class PermissionsViewTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_permissions_endpoint"] = "True"
        return settings


class PermissionsUnauthenticatedViewTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_permissions_endpoint"] = "True"
        settings["bucket_create_principals"] = "system.Everyone"
        return settings

    def setUp(self):
        super().setUp()
        self.everyone_headers = get_user_headers("")
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.everyone_headers)
        self.app.put_json(
            "/buckets/beers/collections/barley",
            MINIMALIST_COLLECTION,
            headers=self.everyone_headers,
        )
        self.app.put_json(
            "/buckets/beers/groups/amateurs", MINIMALIST_GROUP, headers=self.everyone_headers
        )
        self.app.put_json(
            "/buckets/beers/groups/barley", MINIMALIST_GROUP, headers=self.everyone_headers
        )
        self.app.put_json(
            "/buckets/beers/collections/barley/records/{}".format(RECORD_ID),
            MINIMALIST_RECORD,
            headers=self.everyone_headers,
        )

        # Other user.
        self.app.put_json("/buckets/water", MINIMALIST_BUCKET, headers=get_user_headers("alice"))

    def test_permissions_list_entries_for_everyone(self):
        resp = self.app.get("/permissions", headers=self.everyone_headers)
        buckets = resp.json["data"]
        self.assertEqual(len(buckets), 6)

        uris = [bucket["uri"] for bucket in buckets]
        self.assertEqual(
            sorted(uris),
            [
                "/",
                "/buckets/beers",
                "/buckets/beers/collections/barley",
                "/buckets/beers/collections/barley/records/{}".format(RECORD_ID),
                "/buckets/beers/groups/amateurs",
                "/buckets/beers/groups/barley",
            ],
        )

    def test_bucket_create_permission_exists(self):
        resp = self.app.get("/permissions", headers=self.everyone_headers)
        buckets = resp.json["data"]
        toplevel_bucket = [b for b in buckets if b["resource_name"] == "root"]
        self.assertEqual(
            toplevel_bucket,
            [{"permissions": ["bucket:create"], "resource_name": "root", "uri": "/"}],
        )

    def test_permissions_can_be_paginated(self):
        """Verify that pagination doesn't squash same IDs"""
        # kinto_http.js uses this sort of request when listing permissions
        real_permissions = self.app.get("/permissions", headers=self.everyone_headers).json["data"]

        def sort_by_uri(records):
            return sorted(records, key=lambda r: r["uri"])

        for order in ["id", "-id", "uri", "-uri"]:
            for limit in range(1, 10):
                with self.subTest(order=order, limit=limit):
                    records = []
                    resp = self.app.get(
                        "/permissions?_sort={}&_limit={}".format(order, limit),
                        headers=self.everyone_headers,
                    )
                    records = records + resp.json["data"]
                    while "Next-Page" in resp.headers:
                        next_url = resp.headers["Next-Page"][len("http://localhost/v1") :]
                        resp = self.app.get(next_url, headers=self.everyone_headers)
                        records = records + resp.json["data"]

                    self.assertEqual(sort_by_uri(real_permissions), sort_by_uri(records))

    def test_object_details_are_provided(self):
        resp = self.app.get("/permissions", headers=self.everyone_headers)
        entries = resp.json["data"]
        for entry in entries:
            if entry["resource_name"] == "record":
                record_permission = entry
        self.assertEqual(record_permission["record_id"], RECORD_ID)
        self.assertEqual(record_permission["collection_id"], "barley")
        self.assertEqual(record_permission["bucket_id"], "beers")


class EntriesTest(PermissionsViewTest):
    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )
        self.app.put_json("/buckets/beers/groups/amateurs", MINIMALIST_GROUP, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley/records/{}".format(RECORD_ID),
            MINIMALIST_RECORD,
            headers=self.headers,
        )

        # Other user.
        self.app.put_json("/buckets/water", MINIMALIST_BUCKET, headers=get_user_headers("alice"))

    def test_permissions_list_entries_for_current_principal(self):
        resp = self.app.get("/permissions", headers=self.headers)
        permissions = resp.json["data"]
        self.assertEqual(len(permissions), 5)

    def test_permissions_can_be_listed_anonymously(self):
        self.app.patch_json(
            "/buckets/beers/collections/barley",
            {"permissions": {"write": ["system.Everyone"]}},
            headers=self.headers,
        )
        resp = self.app.get("/permissions")
        permissions = resp.json["data"]
        self.assertEqual(len(permissions), 1)

    def test_implicit_permissions_are_explicited(self):
        resp = self.app.get("/permissions", headers=get_user_headers("alice"))
        permissions = resp.json["data"]
        bucket_permissions = permissions[0]
        self.assertEqual(
            sorted(bucket_permissions["permissions"]),
            ["collection:create", "group:create", "read", "read:attributes", "write"],
        )

    def test_object_details_are_provided(self):
        resp = self.app.get("/permissions", headers=self.headers)
        entries = resp.json["data"]
        for entry in entries:
            if entry["resource_name"] == "record":
                record_permission = entry
        self.assertEqual(record_permission["record_id"], RECORD_ID)
        self.assertEqual(record_permission["collection_id"], "barley")
        self.assertEqual(record_permission["bucket_id"], "beers")

    def test_permissions_list_can_be_filtered(self):
        resp = self.app.get(
            "/permissions?in_resource_name=bucket,collection", headers=self.headers
        )
        permissions = resp.json["data"]
        self.assertEqual(len(permissions), 2)

    def test_filtering_with_unknown_field_is_not_supported(self):
        self.app.get("/permissions?movie=bourne", headers=self.headers, status=400)

    def test_permissions_fields_can_be_selected(self):
        resp = self.app.get("/permissions?_fields=uri", headers=self.headers)
        self.assertNotIn("resource_name", resp.json["data"][0])

    def test_permissions_list_can_be_paginated(self):
        resp = self.app.head("/permissions", headers=self.headers)
        self.assertEqual(resp.headers["Total-Objects"], "5")

        resp = self.app.get("/permissions?_limit=2", headers=self.headers)
        self.assertTrue(not resp.headers.get("Total-Objects"))
        self.assertIn("Next-Page", resp.headers)
        self.assertEqual(len(resp.json["data"]), 2)

    def test_permissions_list_do_not_crash_with_preconditions(self):
        headers = {"If-None-Match": '"123"', **self.headers}
        self.app.get("/permissions", headers=headers)

    def test_permissions_can_be_paginated(self):
        for i in range(10):
            self.app.put_json(
                "/buckets/beers/collections/barley/records/r-{}".format(i),
                MINIMALIST_RECORD,
                headers=self.headers,
            )
        resp = self.app.get("/permissions?resource_name=record&_limit=7", headers=self.headers)
        page1 = resp.json["data"]
        next_page = resp.headers["Next-Page"].replace("http://localhost/v1", "")
        resp = self.app.get(next_page, headers=self.headers)
        page2 = resp.json["data"]
        assert len(page1 + page2) == 11  # see setup().

    def test_permissions_can_be_paginated_with_uri_in_sorting(self):
        for i in range(10):
            self.app.put_json(
                "/buckets/beers/collections/barley/records/r-{}".format(i),
                MINIMALIST_RECORD,
                headers=self.headers,
            )
        resp = self.app.get(
            "/permissions?resource_name=record&_limit=7&_sort=uri", headers=self.headers
        )
        page1 = resp.json["data"]
        next_page = resp.headers["Next-Page"].replace("http://localhost/v1", "")
        resp = self.app.get(next_page, headers=self.headers)
        page2 = resp.json["data"]
        assert len(page1 + page2) == 11  # see setup().


class GroupsPermissionTest(PermissionsViewTest):
    def setUp(self):
        super().setUp()

        self.admin_headers = get_user_headers("admin")
        self.admin_principal = self.app.get("/", headers=self.admin_headers).json["user"]["id"]

        self.app.put_json(
            "/buckets/beers",
            {"permissions": {"write": ["/buckets/beers/groups/admins"]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/beers/groups/admins",
            {"data": {"members": [self.admin_principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )

        self.app.put_json("/buckets/sodas", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/groups/admins",
            {"data": {"members": [self.admin_principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/sodas/collections/sprite",
            {"permissions": {"read": ["/buckets/beers/groups/admins"]}},
            headers=self.headers,
        )

    def test_permissions_granted_via_groups_are_listed(self):
        resp = self.app.get("/permissions", headers=self.admin_headers)
        buckets = [e for e in resp.json["data"] if e["resource_name"] == "bucket"]
        self.assertEqual(buckets[0]["id"], "beers")
        self.assertIn("write", buckets[0]["permissions"])

        collections = [e for e in resp.json["data"] if e["resource_name"] == "collection"]
        self.assertEqual(collections[0]["id"], "sprite")
        self.assertEqual(collections[0]["bucket_id"], "sodas")
        self.assertIn("read", collections[0]["permissions"])

    def test_permissions_inherited_are_not_listed(self):
        resp = self.app.get("/permissions", headers=self.admin_headers)
        collections = [
            e
            for e in resp.json["data"]
            # top level '/' does not have an id
            if e["uri"] != "/"
            if e["bucket_id"] == "beers" and e["resource_name"] == "collection"
        ]
        self.assertEqual(len(collections), 0)


class SettingsPermissionsTest(PermissionsViewTest):
    admin_headers = get_user_headers("admin")
    admin_principal = "basicauth:bb7fe7b98e759578ef0de85b546dd57d21fe1e399390ad8dafc9886043a00e5c"  # NOQA

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["bucket_write_principals"] = "system.Authenticated"
        settings["group_create_principals"] = cls.admin_principal
        settings["collection_write_principals"] = "system.Authenticated"
        settings["record_create_principals"] = "/buckets/beers/groups/admins"
        return settings

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/groups/admins",
            {"data": {"members": [self.admin_principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )

    def test_bucket_write_taken_into_account(self):
        resp = self.app.get("/permissions", headers=get_user_headers("any"))
        buckets = [e for e in resp.json["data"] if e["resource_name"] == "bucket"]
        self.assertEqual(buckets[0]["id"], "beers")
        self.assertIn("write", buckets[0]["permissions"])

    def test_collection_create_taken_into_account(self):
        resp = self.app.get("/permissions", headers=self.admin_headers)
        buckets = [e for e in resp.json["data"] if e["resource_name"] == "bucket"]
        self.assertEqual(buckets[0]["id"], "beers")
        self.assertIn("group:create", buckets[0]["permissions"])

    def test_collection_write_taken_into_account(self):
        resp = self.app.get("/permissions", headers=get_user_headers("any"))
        collections = [e for e in resp.json["data"] if e["resource_name"] == "collection"]
        self.assertEqual(collections[0]["id"], "barley")
        self.assertIn("write", collections[0]["permissions"])

    def test_record_create_taken_into_account(self):
        resp = self.app.get("/permissions", headers=self.admin_headers)
        collections = [e for e in resp.json["data"] if e["resource_name"] == "collection"]
        self.assertEqual(collections[0]["id"], "barley")
        self.assertIn("record:create", collections[0]["permissions"])

    def test_settings_permissions_are_merged_with_perms_backend(self):
        self.app.patch_json(
            "/buckets/beers",
            {"permissions": {"collection:create": [self.admin_principal]}},
            headers=self.headers,
        )
        self.app.patch_json(
            "/buckets/beers/collections/barley",
            {"permissions": {"read": [self.admin_principal]}},
            headers=self.headers,
        )

        resp = self.app.get("/permissions", headers=self.admin_headers)

        buckets = [e for e in resp.json["data"] if e["resource_name"] == "bucket"]
        self.assertEqual(buckets[0]["id"], "beers")
        self.assertIn("group:create", buckets[0]["permissions"])
        self.assertIn("collection:create", buckets[0]["permissions"])

        collections = [e for e in resp.json["data"] if e["resource_name"] == "collection"]
        self.assertEqual(collections[0]["id"], "barley")
        self.assertIn("record:create", collections[0]["permissions"])
        self.assertIn("read", collections[0]["permissions"])


class DeletedObjectsTest(PermissionsViewTest):
    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json("/buckets/beers/groups/admins", MINIMALIST_GROUP, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )
        self.app.put_json(
            "/buckets/beers/collections/barley/records/vieuxsinge",
            MINIMALIST_RECORD,
            headers=self.headers,
        )
        self.app.delete("/buckets/beers", headers=self.headers)

    def test_deleted_objects_are_not_listed(self):
        resp = self.app.get("/permissions", headers=self.headers)
        self.assertEqual(len(resp.json["data"]), 1)


class IgnoredSettingsTest(PermissionsViewTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["signer.auto_create_resources_principals"] = "system.Authenticated"
        return settings

    def test_unrecognized_settings_are_ignored(self):
        resp = self.app.get("/permissions", headers=self.headers)
        self.assertEqual(
            resp.json["data"],
            [{"permissions": ["bucket:create"], "resource_name": "root", "uri": "/"}],
        )



import time
import unittest
from unittest import mock

from kinto.core.cache import CacheBase
from kinto.core.cache import memcached as memcached_backend
from kinto.core.cache import memory as memory_backend
from kinto.core.cache import postgresql as postgresql_backend
from kinto.core.cache.testing import CacheTest
from kinto.core.testing import skip_if_no_memcached, skip_if_no_postgresql
from kinto.core.utils import memcache, sqlalchemy


class CacheBaseTest(unittest.TestCase):
    def setUp(self):
        self.cache = CacheBase(cache_prefix="")

    def test_mandatory_overrides(self):
        calls = [
            (self.cache.initialize_schema,),
            (self.cache.flush,),
            (self.cache.ttl, ""),
            (self.cache.expire, "", ""),
            (self.cache.get, ""),
            (self.cache.set, "", "", 42),
            (self.cache.delete, ""),
        ]
        for call in calls:
            self.assertRaises(NotImplementedError, *call)


class MemoryCacheTest(CacheTest, unittest.TestCase):
    backend = memory_backend
    settings = {"cache_prefix": "", "cache_max_size_bytes": 7000}

    def get_backend_prefix(self, prefix):
        backend_prefix = CacheTest.get_backend_prefix(self, prefix)

        # Share the store between both client for tests.
        backend_prefix._ttl = self.cache._ttl
        backend_prefix._store = self.cache._store

        return backend_prefix

    def test_backend_error_is_raised_anywhere(self):
        pass

    def test_ping_returns_false_if_unavailable(self):
        pass

    def test_ping_logs_error_if_unavailable(self):
        pass

    def test_clean_expired_expires_items(self):
        self.cache.set("foobar", "toto", 0.01)
        assert "foobar" in self.cache._store
        assert "foobar" in self.cache._ttl
        assert "foobar" in self.cache._created_at
        time.sleep(0.02)
        retrieved = self.cache._clean_expired()
        assert "foobar" not in self.cache._store
        assert "foobar" not in self.cache._ttl
        assert "foobar" not in self.cache._created_at
        self.assertIsNone(retrieved)

    def test_add_over_quota_clean_oversized_items(self):
        for x in range(100):
            # Each entry is 70 bytes
            self.cache.set("foo{0:03d}".format(x), "toto", 42)
            time.sleep(0.001)
        assert self.cache.get("foo000") == "toto"
        # This should delete the 2 first entries
        self.cache.set("foobar", "tata", 42)
        assert self.cache._quota == 7000 - 70 * 20
        assert self.cache.get("foo000") is None
        assert self.cache.get("foobar") == "tata"

    def test_size_quota_can_be_set_to_zero(self):
        before = self.cache.max_size_bytes
        self.cache.max_size_bytes = 0
        self.cache.set("foobar", "tata", 42)
        self.cache.max_size_bytes = before
        assert self.cache.get("foobar") == "tata"


@skip_if_no_memcached
class MemcachedCacheTest(CacheTest, unittest.TestCase):
    backend = memcached_backend
    settings = {"cache_prefix": "", "cache_hosts": "127.0.0.1:11211"}

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.cache._client.servers[0], "connect", side_effect=memcache.Client.MemcachedKeyError
        )

    def test_set_with_ttl_expires_the_value(self):
        self.cache.set("foobar", "toto", 1)
        time.sleep(1.1)
        retrieved = self.cache.get("foobar")
        self.assertIsNone(retrieved)

    def test_expire_expires_the_value(self):
        self.cache.set("foobar", "toto", 42)
        self.cache.expire("foobar", 1)
        time.sleep(1.1)
        retrieved = self.cache.get("foobar")
        self.assertIsNone(retrieved)


@skip_if_no_postgresql
class PostgreSQLCacheTest(CacheTest, unittest.TestCase):
    backend = postgresql_backend
    settings = {
        "cache_backend": "kinto.core.cache.postgresql",
        "cache_pool_size": 10,
        "cache_url": "postgresql://postgres:postgres@localhost:5432/testdb",
        "cache_prefix": "",
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.cache.client, "session_factory", side_effect=sqlalchemy.exc.SQLAlchemyError
        )



from unittest import mock

from pyramid.request import Request

from kinto.core import utils
from kinto.core.authorization import AuthorizationPolicy, RouteFactory, groupfinder
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.testing import DummyRequest, unittest


class RouteFactoryTest(unittest.TestCase):
    def setUp(self):
        self.object_uri = "/foo/bar"

    def assert_request_resolves_to(self, method, permission, uri=None, object_not_found=False):
        if uri is None:
            uri = self.object_uri

        with mock.patch("kinto.core.utils.current_service") as current_service:
            # Patch current service.
            resource = mock.MagicMock()
            resource.object_id = 1
            if object_not_found:
                resource.model.get_object.side_effect = storage_exceptions.ObjectNotFoundError
            else:
                resource.model.get_object.return_value = 1
            current_service().resource.return_value = resource

            # Do the actual call.
            request = DummyRequest(method=method)
            request.upath_info = uri
            context = RouteFactory(request)

            self.assertEqual(context.required_permission, permission)

    def test_http_unknown_does_not_raise_a_500(self):
        self.assert_request_resolves_to("unknown", None)

    def test_http_get_resolves_in_a_read_permission(self):
        self.assert_request_resolves_to("get", "read")

    def test_http_post_resolves_in_a_create_permission(self):
        self.assert_request_resolves_to("post", "create")

    def test_http_delete_resolves_in_a_write_permission(self):
        self.assert_request_resolves_to("delete", "write")

    def test_http_put_unexisting_object_resolves_in_a_create_permission(self):
        with mock.patch("kinto.core.utils.current_service") as current_service:
            # Patch current service.
            resource = mock.MagicMock()
            resource.object_id = 1
            resource.model.get_object.side_effect = storage_exceptions.ObjectNotFoundError
            current_service().resource.return_value = resource
            current_service().plural_path = "/school/{school_id}"
            # Do the actual call.
            request = DummyRequest(method="put")
            request.upath_info = "/school/abc/students/1"
            request.matchdict = {"school_id": "abc"}
            context = RouteFactory(request)

            self.assertEqual(context.required_permission, "create")

    def test_http_put_existing_object_resolves_in_a_write_permission(self):
        self.assert_request_resolves_to("put", "write")

    def test_http_put_sets_current_object_attribute(self):
        with mock.patch("kinto.core.utils.current_service") as current_service:
            # Patch current service.
            resource = mock.MagicMock()
            resource.object_id = 1
            resource.model.get_object.return_value = mock.sentinel.object
            current_service().resource.return_value = resource
            # Do the actual call.
            request = DummyRequest(method="put")
            context = RouteFactory(request)
            self.assertEqual(context.current_object, mock.sentinel.object)

    def test_http_patch_resolves_in_a_write_permission(self):
        self.assert_request_resolves_to("patch", "write")

    def test_attributes_are_none_with_blank_requests(self):
        request = Request.blank(path="/")
        request.registry = mock.Mock(settings={})
        request.authn_type = "fxa"
        request.prefixed_userid = property(utils.prefixed_userid)
        context = RouteFactory(request)
        self.assertIsNone(context.required_permission)
        self.assertIsNone(context.current_object)
        self.assertIsNone(context.resource_name)

    def test_attributes_are_none_with_non_resource_requests(self):
        basic_service = object()
        request = Request.blank(path="/")
        request.prefixed_userid = property(utils.prefixed_userid)
        request.matched_route = mock.Mock(pattern="foo")
        request.registry = mock.Mock(cornice_services={"foo": basic_service})
        request.registry.settings = {}

        context = RouteFactory(request)
        self.assertIsNone(context.current_object)
        self.assertIsNone(context.required_permission)
        self.assertIsNone(context.resource_name)

    def test_fetch_shared_objects_uses_pattern_if_on_plural_endpoint(self):
        request = DummyRequest()
        request.route_path.return_value = "/v1/buckets/%2A"
        service = mock.MagicMock()
        service.type = "plural"
        with mock.patch("kinto.core.authorization.utils.current_service") as m:
            m.return_value = service
            context = RouteFactory(request)
        self.assertTrue(context.on_plural_endpoint)

        context.fetch_shared_objects("read", ["userid"], None)

        request.registry.permission.get_accessible_objects.assert_called_with(
            ["userid"], [("/buckets/*", "read")], with_children=False
        )

    def test_fetch_shared_objects_uses_get_bound_permission_callback(self):
        request = DummyRequest()
        service = mock.MagicMock()
        request.route_path.return_value = "/v1/buckets/%2A"
        service.type = "plural"
        with mock.patch("kinto.core.authorization.utils.current_service") as m:
            m.return_value = service
            context = RouteFactory(request)
        self.assertTrue(context.on_plural_endpoint)

        # Define a callback where write means read:
        def get_bound_perms(obj_id, perm):
            return [(obj_id, "write"), (obj_id, "read")]

        context.fetch_shared_objects("read", ["userid"], get_bound_perms)

        request.registry.permission.get_accessible_objects.assert_called_with(
            ["userid"], [("/buckets/*", "write"), ("/buckets/*", "read")], with_children=False
        )

    def test_fetch_shared_objects_sets_shared_ids_from_results(self):
        request = DummyRequest()
        context = RouteFactory(request)
        request.registry.permission.get_accessible_objects.return_value = {
            "/obj/1": ["read", "write"],
            "/obj/3": ["obj:create"],
        }
        context.fetch_shared_objects("read", ["userid"], None)
        self.assertEqual(sorted(context.shared_ids), ["1", "3"])

    def test_fetch_shared_objects_sets_shared_ids_if_empty(self):
        request = DummyRequest()
        context = RouteFactory(request)
        request.registry.permission.get_accessible_objects.return_value = {}

        context.fetch_shared_objects("read", ["userid"], None)

        self.assertEqual(context.shared_ids, [])

    def test_permits_takes_route_factory_allowed_principals_into_account_for_object_creation(self):
        request = DummyRequest()
        context = RouteFactory(request)
        context._check_permission.return_value = False
        context.resource_name = "book"
        context.required_permission = "book:create"
        context._settings = {"book_create_principals": "fxa:user"}
        self.assertTrue(context.check_permission(["fxa:user"], None))


class AuthorizationPolicyTest(unittest.TestCase):
    def setUp(self):
        self.authz = AuthorizationPolicy()
        self.context = mock.MagicMock()
        self.context.permission_object_id = "/articles/43/comments/2"
        self.context.required_permission = "read"
        self.principals = ["portier:yourself"]
        self.context.get_prefixed_principals.return_value = self.principals
        self.permission = "dynamic"

    def test_permits_logs_authz_failures(self):
        self.context.on_plural_endpoint = False
        self.context.check_permission.return_value = False
        with mock.patch("kinto.core.authorization.logger") as mocked:
            self.authz.permits(self.context, self.principals, "dynamic")
        userid = "portier:yourself"
        object_id = "/articles/43/comments/2"
        perm = "read"
        mocked.info.assert_called_with(
            "Permission %r on %r not granted to %r.",
            perm,
            object_id,
            userid,
            extra=dict(perm=perm, uri=object_id, userid=userid),
        )

    def test_permits_refers_to_context_to_check_permissions(self):
        self.context.check_permission.return_value = True
        allowed = self.authz.permits(self.context, self.principals, "dynamic")
        self.assertTrue(allowed)

    def test_permits_refers_to_context_to_check_permission_principals(self):
        self.context.check_permission.side_effect = (False, True)
        self.context.current_object = None
        self.context.resource_name = "object"
        self.context.fetch_shared_objects = mock.MagicMock(return_value=[])

        allowed = self.authz.permits(self.context, self.principals, "dynamic")

        self.assertTrue(allowed)
        self.context.check_permission.assert_called_with(
            self.principals, [("/articles/43/comments", "object:create")]
        )

    def test_permits_reads_the_context(self):
        self.authz.permits(self.context, self.principals, "dynamic")
        self.context.check_permission.assert_called_with(
            self.principals,
            [
                (self.context.permission_object_id, "read"),
                (self.context.permission_object_id, "write"),
            ],
        )

    def test_permits_bypasses_the_permissions_backend_if_private(self):
        self.authz.permits(self.context, self.principals, "private")
        self.context.check_permission.assert_not_called()

    def test_permits_returns_true_if_authenticated(self):
        assert self.authz.permits(self.context, ["system.Authenticated"], "private")
        assert not self.authz.permits(self.context, [], "private")

    def test_permits_uses_get_bound_permissions_if_defined(self):
        self.authz.get_bound_permissions = lambda o, p: mock.sentinel.callback
        self.authz.permits(self.context, self.principals, "dynamic")
        self.context.check_permission.assert_called_with(self.principals, mock.sentinel.callback)

    def test_permits_calls_get_bound_permissions_with_context_info(self):
        self.authz.get_bound_permissions = mock.Mock(return_value=[])
        self.authz.permits(self.context, self.principals, "dynamic")
        self.authz.get_bound_permissions.assert_called_with(
            self.context.permission_object_id, "read"
        )

    def test_permits_prepend_obj_type_to_permission_on_create(self):
        self.context.required_permission = "create"
        self.context.resource_name = "object"
        self.authz.permits(self.context, self.principals, "dynamic")
        self.context.check_permission.assert_called_with(
            self.principals, [(self.context.permission_object_id, "object:create")]
        )

    def test_permits_takes_route_factory_allowed_principals_into_account(self):
        self.context.resource_name = "object"
        self.context.required_permission = "create"
        self.context._settings = {"object_create_principals": "fxa:user"}
        allowed = self.authz.permits(self.context, self.principals, "dynamic")
        self.context._check_permission.assert_not_called()
        self.assertTrue(allowed)


class GuestAuthorizationPolicyTest(unittest.TestCase):
    def setUp(self):
        self.authz = AuthorizationPolicy()
        self.authz.get_bound_permissions = lambda o, p: []
        self.request = DummyRequest(method="GET")
        self.context = RouteFactory(self.request)
        self.context.on_plural_endpoint = True
        self.context.check_permission = mock.Mock(return_value=False)

    def test_permits_returns_true_if_plural_endpoint_and_shared_objects(self):
        self.context.fetch_shared_objects = mock.MagicMock(return_value=["object1", "object2"])
        allowed = self.authz.permits(self.context, ["userid"], "dynamic")
        # Note: we use the list of principals from request.prefixed_principals
        self.context.fetch_shared_objects.assert_called_with(
            "read",
            ["basicauth:bob", "system.Everyone", "system.Authenticated"],
            self.authz.get_bound_permissions,
        )
        self.assertTrue(allowed)

    def test_permits_does_not_return_true_if_not_plural_endpoint(self):
        self.context.on_plural_endpoint = False
        allowed = self.authz.permits(self.context, ["userid"], "dynamic")
        self.assertFalse(allowed)

    def test_permits_does_not_return_true_if_not_list_operation(self):
        self.context.required_permission = "create"
        allowed = self.authz.permits(self.context, ["userid"], "dynamic")
        self.assertFalse(allowed)
        allowed = self.authz.permits(self.context, ["userid"], "create")
        self.assertFalse(allowed)

    def test_permits_returns_false_if_resource_is_unknown(self):
        self.context.fetch_shared_objects = mock.MagicMock(return_value=None)
        allowed = self.authz.permits(self.context, ["userid"], "dynamic")
        # Note: we use the list of principals from request.prefixed_principals
        self.context.fetch_shared_objects.assert_called_with(
            "read",
            ["basicauth:bob", "system.Everyone", "system.Authenticated"],
            self.authz.get_bound_permissions,
        )
        self.assertFalse(allowed)

    def test_perm_object_id_is_naive_if_no_object_path_exists(self):
        def route_path(service_name, **kwargs):
            # Simulate a resource that has no object_path (only list).
            if service_name == "article-object":
                raise KeyError
            return "/comments/sub/{id}".format_map(kwargs)

        self.request.route_path.side_effect = route_path

        self.request.path = "/comments"
        self.context.resource_name = "comment"
        obj_id = self.context.get_permission_object_id(self.request, "*")
        self.assertEqual(obj_id, "/comments/sub/*")

        self.request.path = "/articles"
        self.context.resource_name = "article"
        obj_id = self.context.get_permission_object_id(self.request, "*")
        self.assertEqual(obj_id, "/articles/*")


class GroupFinderTest(unittest.TestCase):
    def setUp(self):
        self.request = DummyRequest(method="GET")

    def test_uses_prefixed_as_userid(self):
        self.request.prefixed_userid = "basic:bob"
        groupfinder("bob", self.request)
        self.request.registry.permission.get_user_principals.assert_called_with("basic:bob")

    def test_uses_provided_id_if_no_prefixed_userid(self):
        self.request.prefixed_userid = None
        groupfinder("bob", self.request)
        self.request.registry.permission.get_user_principals.assert_called_with("bob")



import colander
from pyramid.authorization import Authenticated
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core import Service


hello = Service(name="hello", path="/", description="Welcome")


class HelloResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown="preserve"))


hello_response_schemas = {
    "200": HelloResponseSchema(description="Return information about the running Instance.")
}


@hello.get(
    permission=NO_PERMISSION_REQUIRED,
    tags=["Utilities"],
    operation_id="server_info",
    response_schemas=hello_response_schemas,
)
def get_hello(request):
    """Return information regarding the current instance."""
    settings = request.registry.settings
    project_name = settings["project_name"]
    project_version = settings["project_version"]
    data = dict(
        project_name=project_name,
        project_version=project_version,
        http_api_version=settings["http_api_version"],
        project_docs=settings["project_docs"],
        url=request.route_url(hello.name),
    )

    eos = get_eos(request)
    if eos:
        data["eos"] = eos

    data["settings"] = {}
    public_settings = request.registry.public_settings
    for setting in list(public_settings):
        data["settings"][setting] = settings[setting]

    # If current user is authenticated, add user info:
    # (Note: this will call authenticated_userid() with multiauth+groupfinder)
    if Authenticated in request.effective_principals:
        data["user"] = request.get_user_info()

    if settings["readonly"]:
        # Information can be cached.
        cache_seconds = int(settings["root_cache_expires_seconds"])
        request.response.cache_expires(cache_seconds)

    # Application can register and expose arbitrary capabilities.
    data["capabilities"] = request.registry.api_capabilities

    return data


def get_eos(request):
    return request.registry.settings["eos"]



from kinto.core import Service
from kinto.core.authorization import RouteFactory


attachment = Service(
    name="attachment",
    description="Attach file to object",
    path="/attachment",
    factory=RouteFactory,
)


@attachment.post(permission="attach")
def attachment_post(request):
    return {"ok": True}


log = Service(
    name="log",
    description="Test endpoint without permissions",
    path="/log",
)


@log.get()
def log_get(request):
    return {}


@log.post()
def log_post(request):
    return {}


@log.delete()
def log_delete(request):
    return {}


@log.put()
def log_put(request):
    return {}


@log.patch()
def log_patch(request):
    return {}


def includeme(config):
    config.add_cornice_service(attachment)
    config.add_cornice_service(log)



import json
import re
import unittest
from unittest import mock

from kinto.core.testing import get_user_headers

from .support import (
    MINIMALIST_BUCKET,
    MINIMALIST_COLLECTION,
    MINIMALIST_GROUP,
    MINIMALIST_RECORD,
    BaseWebTest,
)


class RecordsViewTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/collections/barley/records"
    _record_url = "/buckets/beers/collections/barley/records/{}"

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )
        resp = self.app.post_json(self.collection_url, MINIMALIST_RECORD, headers=self.headers)
        self.record = resp.json["data"]
        self.record_url = self._record_url.format(self.record["id"])

    def test_records_can_be_accessed_by_id(self):
        self.app.get(self.record_url, headers=self.headers)

    def test_unknown_bucket_raises_403(self):
        other_bucket = self.collection_url.replace("beers", "sodas")
        self.app.get(other_bucket, headers=self.headers, status=403)

    def test_unknown_collection_raises_404(self):
        other_collection = self.collection_url.replace("barley", "pills")
        resp = self.app.get(other_collection, headers=self.headers, status=404)
        self.assertEqual(resp.json["details"]["id"], "pills")
        self.assertEqual(resp.json["details"]["resource_name"], "collection")

    def test_unknown_record_raises_404(self):
        other_record = self.record_url.replace(self.record["id"], self.record["id"] + "blah")
        response = self.app.get(other_record, headers=self.headers, status=404)
        self.assertEqual(response.json["details"]["id"], self.record["id"] + "blah")
        self.assertEqual(response.json["details"]["resource_name"], "record")

    def test_unknown_collection_does_not_query_timestamp(self):
        other_collection = self.collection_url.replace("barley", "pills")
        patch = mock.patch.object(self.app.app.registry.storage, "resource_timestamp")
        self.addCleanup(patch.stop)
        mocked = patch.start()
        self.app.get(other_collection, headers=self.headers, status=404)
        self.assertFalse(mocked.called)

    def test_parent_collection_and_bucket_are_fetched_only_once_in_batch(self):
        batch = {"requests": []}
        nb_create = 25
        for i in range(nb_create):
            request = {"method": "POST", "path": self.collection_url, "body": MINIMALIST_RECORD}
            batch["requests"].append(request)

        with mock.patch.object(self.storage, "get", wraps=self.storage.get) as patched:
            self.app.post_json("/batch", batch, headers=self.headers)
            self.assertEqual(patched.call_count, 1)

    def test_individual_collections_can_be_deleted(self):
        resp = self.app.get(self.collection_url, headers=self.headers)
        self.assertEqual(len(resp.json["data"]), 1)
        self.app.delete(self.collection_url, headers=self.headers)
        resp = self.app.get(self.collection_url, headers=self.headers)
        self.assertEqual(len(resp.json["data"]), 0)

    def test_records_can_be_added_to_collections(self):
        response = self.app.get(self.record_url, headers=self.headers)
        record = response.json["data"]
        del record["id"]
        del record["last_modified"]
        self.assertEqual(record, MINIMALIST_RECORD["data"])

    def test_records_are_isolated_by_bucket_and_by_collection(self):
        # By collection.
        self.app.put_json(
            "/buckets/beers/collections/pills", MINIMALIST_BUCKET, headers=self.headers
        )
        other_collection = self.record_url.replace("barley", "pills")
        self.app.get(other_collection, headers=self.headers, status=404)

        # By bucket.
        self.app.put_json("/buckets/sodas", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/sodas/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )
        other_bucket = self.record_url.replace("beers", "sodas")
        self.app.get(other_bucket, headers=self.headers, status=404)

        # By bucket and by collection.
        self.app.put_json("/buckets/be", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/be/collections/ba", MINIMALIST_COLLECTION, headers=self.headers
        )
        other = self.record_url.replace("barley", "ba").replace("beers", "be")
        self.app.get(other, headers=self.headers, status=404)

    def test_a_collection_named_group_do_not_interfere_with_groups(self):
        # Create a group.
        self.app.put_json("/buckets/beers/groups/test", MINIMALIST_GROUP, headers=self.headers)
        # Create a record in a collection named "group".
        self.app.put_json(
            "/buckets/beers/collections/groups", MINIMALIST_COLLECTION, headers=self.headers
        )
        collection_group = self.collection_url.replace("barley", "groups")
        self.app.post_json(collection_group, MINIMALIST_RECORD, headers=self.headers)
        # There is still only one group.
        resp = self.app.get("/buckets/beers/groups", headers=self.headers)
        self.assertEqual(len(resp.json["data"]), 1)

    def test_records_can_be_filtered_on_any_field(self):
        self.app.post_json(self.collection_url, MINIMALIST_RECORD, headers=self.headers)
        response = self.app.get(self.collection_url + "?unknown=1", headers=self.headers)
        self.assertEqual(len(response.json["data"]), 0)

    def test_records_can_be_sorted_on_any_field(self):
        for i in range(3):
            record = {
                **MINIMALIST_RECORD,
                "data": {**MINIMALIST_RECORD["data"], "name": "Stout {}".format(i)},
            }
            self.app.post_json(self.collection_url, record, headers=self.headers)

        response = self.app.get(self.collection_url + "?_sort=-name", headers=self.headers)
        names = [i["name"] for i in response.json["data"]]
        self.assertEqual(names, ["Stout 2", "Stout 1", "Stout 0", "Hulled Barley"])

    def test_wrong_create_permissions_cannot_be_added_on_records(self):
        record = {**MINIMALIST_RECORD, "permissions": {"record:create": ["fxa:user"]}}
        self.app.put_json(self.record_url, record, headers=self.headers, status=400)

    def test_create_a_record_update_resource_timestamp(self):
        collection_resp = self.app.get(self.collection_url, headers=self.headers)
        old_timestamp = int(json.loads(collection_resp.headers["ETag"]))
        self.app.post_json(
            self.collection_url, MINIMALIST_RECORD, headers=self.headers, status=201
        )
        collection_resp = self.app.get(self.collection_url, headers=self.headers)
        new_timestamp = int(json.loads(collection_resp.headers["ETag"]))
        assert old_timestamp < new_timestamp

    def test_create_a_record_without_id_generates_a_uuid(self):
        resp = self.app.post_json(
            self.collection_url, MINIMALIST_RECORD, headers=self.headers, status=201
        )
        regexp = re.compile(r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-" r"[0-9a-f]{4}-[0-9a-f]{12}$")
        self.assertTrue(regexp.match(resp.json["data"]["id"]))

    def test_create_a_record_with_an_id_uses_it(self):
        record = {"data": dict(id="a-simple-id", **MINIMALIST_RECORD["data"])}
        resp = self.app.post_json(self.collection_url, record, headers=self.headers, status=201)
        self.assertEqual(resp.json["data"]["id"], "a-simple-id")

    def test_create_a_record_with_an_existing_id_returns_existing(self):
        resp = self.app.post_json(
            self.collection_url, MINIMALIST_RECORD, headers=self.headers, status=201
        )
        existing_id = resp.json["data"]["id"]
        record = {"data": {"id": existing_id, "stars": 8}}
        resp = self.app.post_json(self.collection_url, record, headers=self.headers, status=200)
        self.assertNotIn("stars", resp.json["data"])

    def test_create_a_record_with_existing_from_someone_else_gives_403(self):
        resp = self.app.post_json(
            self.collection_url, MINIMALIST_RECORD, headers=self.headers, status=201
        )
        existing_id = resp.json["data"]["id"]
        record = {"data": {"id": existing_id, "stars": 8}}
        resp = self.app.post_json(
            self.collection_url, record, headers=get_user_headers("tartanpion"), status=403
        )

    def test_update_a_record_update_resource_timestamp(self):
        collection_resp = self.app.get(self.collection_url, headers=self.headers)
        old_timestamp = int(json.loads(collection_resp.headers["ETag"]))
        self.app.put_json(self.record_url, MINIMALIST_RECORD, headers=self.headers, status=200)
        collection_resp = self.app.get(self.collection_url, headers=self.headers)
        new_timestamp = int(json.loads(collection_resp.headers["ETag"]))
        assert old_timestamp < new_timestamp

    def test_delete_a_record_update_resource_timestamp(self):
        collection_resp = self.app.get(self.collection_url, headers=self.headers)
        old_timestamp = int(json.loads(collection_resp.headers["ETag"]))
        self.app.delete(self.record_url, headers=self.headers, status=200)
        collection_resp = self.app.get(self.collection_url, headers=self.headers)
        new_timestamp = int(json.loads(collection_resp.headers["ETag"]))
        assert old_timestamp < new_timestamp

    def test_record_is_accessible_by_group_member(self):
        # access as aaron
        self.aaron_headers = {**self.headers, **get_user_headers("aaron")}

        resp = self.app.get("/", headers=self.aaron_headers, status=200)

        self.create_group("beers", "brewers", [resp.json["user"]["id"]])
        record = {**MINIMALIST_RECORD, "permissions": {"read": ["/buckets/beers/groups/brewers"]}}
        self.app.put_json(self.record_url, record, headers=self.headers, status=200)

        self.app.get(self.record_url, headers=self.aaron_headers, status=200)

    def test_records_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, "Content-Type": "text/plain"}
        self.app.put(self.record_url, MINIMALIST_RECORD, headers=headers, status=415)

    def test_records_should_reject_unaccepted_client_accept(self):
        headers = {**self.headers, "Accept": "text/plain"}
        self.app.get(self.record_url, MINIMALIST_RECORD, headers=headers, status=406)

    def test_records_should_accept_client_accept(self):
        headers = {**self.headers, "Accept": "*/*"}
        self.app.get(self.record_url, MINIMALIST_RECORD, headers=headers, status=200)

    def test_records_can_be_created_after_deletion(self):
        self.app.delete(self.record_url, headers=self.headers, status=200)
        headers = {**self.headers, "If-None-Match": "*"}
        self.app.put_json(self.record_url, MINIMALIST_RECORD, headers=headers, status=201)


class RecordsViewMergeTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/collections/barley/records"
    _record_url = "/buckets/beers/collections/barley/records/{}"

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )
        record = {**MINIMALIST_RECORD, "data": {"grain": {"one": 1}}}
        resp = self.app.post_json(self.collection_url, record, headers=self.headers)
        self.record = resp.json["data"]
        self.record_url = self._record_url.format(self.record["id"])

    def test_merge_patch(self):
        headers = {**self.headers, "Content-Type": "application/merge-patch+json"}
        json = {"data": {"grain": {"two": 2}}}
        resp = self.app.patch_json(self.record_url, json, headers=headers, status=200)
        self.assertEqual(resp.json["data"]["grain"]["one"], 1)
        self.assertEqual(resp.json["data"]["grain"]["two"], 2)

    def test_merge_patch_remove_nones(self):
        headers = {**self.headers, "Content-Type": "application/merge-patch+json"}
        json = {"data": {"grain": {"one": None}}}
        resp = self.app.patch_json(self.record_url, json, headers=headers, status=200)
        self.assertNotIn("one", resp.json["data"]["grain"])

    def test_merge_patch_permissions(self):
        headers = {**self.headers, "Content-Type": "application/merge-patch+json"}

        # Add perms
        json = {"permissions": {"read": ["bad_guys"], "write": ["good_guys"]}}
        resp = self.app.patch_json(self.record_url, json, headers=headers, status=200)
        result = resp.json["permissions"]

        self.assertIn("write", result)
        self.assertIn("read", result)
        self.assertIn("bad_guys", result["read"])
        self.assertIn("good_guys", result["write"])

        # Remove perms
        resp = self.app.patch_json(self.record_url, json, headers=headers, status=200)

        json = {"permissions": {"read": None, "write": ["good_guys"]}}
        resp = self.app.patch_json(self.record_url, json, headers=headers, status=200)
        result = resp.json["permissions"]

        self.assertIn("write", result)
        self.assertNotIn("read", result)
        self.assertIn("good_guys", result["write"])


class RecordsViewPatchTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/collections/barley/records"
    _record_url = "/buckets/beers/collections/barley/records/{}"

    def setUp(self):
        super().setUp()
        self.patch_headers = {**self.headers, "Content-Type": "application/json-patch+json"}

        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )
        record = {
            **MINIMALIST_RECORD,
            "permissions": {"read": ["alice", "carla"], "write": ["bob"]},
        }
        resp = self.app.post_json(self.collection_url, record, headers=self.headers)
        self.record = resp.json["data"]
        self.record_url = self._record_url.format(self.record["id"])

    def test_patch_add_permissions(self):
        json = [{"op": "add", "path": "/permissions/read/me", "value": "me"}]
        resp = self.app.patch_json(self.record_url, json, headers=self.patch_headers, status=200)

        perms = resp.json["permissions"]
        self.assertIn("me", perms["read"])
        self.assertIn("alice", perms["read"])

    def test_patch_update_permissions(self):
        json = [{"op": "add", "path": "/permissions/read", "value": ["me"]}]
        resp = self.app.patch_json(self.record_url, json, headers=self.patch_headers, status=200)

        perms = resp.json["permissions"]
        self.assertIn("me", perms["read"])
        self.assertNotIn(("alice", "bob"), perms["read"])

    def test_patch_remove_permissions(self):
        json = [{"op": "remove", "path": "/permissions/read/alice"}]
        resp = self.app.patch_json(self.record_url, json, headers=self.patch_headers, status=200)
        perms = resp.json["permissions"]
        self.assertNotIn("alice", perms["read"])

    def test_patch_move_permissions(self):
        json = [{"op": "move", "from": "/permissions/read/alice", "path": "/data/old"}]
        resp = self.app.patch_json(self.record_url, json, headers=self.patch_headers, status=200)

        perms = resp.json["permissions"]
        data = resp.json["data"]
        self.assertNotIn("alice", perms["read"])
        self.assertEqual("alice", data["old"])

    def test_patch_raises_400_on_wrong_path(self):
        json = [{"op": "add", "path": "/permissions/destroy/me"}]
        self.app.patch_json(self.record_url, json, headers=self.patch_headers, status=400)


class RecordsViewFilterTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/collections/barley/records"

    RECORDS = [
        {
            "id": "strawberry",
            "flavor": "strawberry",
            "attributes": {"ibu": 25, "seen_on": "2017-06-01"},
            "author": None,
        },
        {"id": "raspberry-1", "flavor": "raspberry", "attributes": {}},
        {"id": "raspberry-2", "flavor": "raspberry", "attributes": []},
        {
            "id": "raspberry-3",
            "flavor": "raspberry",
            "attributes": {"ibu": 25, "seen_on": "2017-06-01", "price": 9.99},
        },
    ]

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )

        for record in self.RECORDS:
            self.app.post_json(self.collection_url, {"data": record}, headers=self.headers)

    def test_records_can_be_filtered_using_json(self):
        response = self.app.get(self.collection_url + '?flavor="strawberry"', headers=self.headers)
        assert len(response.json["data"]) == 1
        assert response.json["data"][0]["id"] == "strawberry"

    def test_records_can_be_filtered_with_object(self):
        query = self.collection_url + '?attributes={"ibu": 25, "seen_on": "2017-06-01"}'
        response = self.app.get(query, headers=self.headers)
        assert len(response.json["data"]) == 1
        assert response.json["data"][0]["id"] == "strawberry"



import uuid
from contextlib import contextmanager
from unittest import mock

from pyramid.config import Configurator

from kinto.core import statsd
from kinto.core.events import (
    ACTIONS,
    AfterResourceChanged,
    AfterResourceRead,
    ResourceChanged,
    ResourceRead,
    notify_resource_event,
)
from kinto.core.storage.exceptions import BackendError
from kinto.core.testing import unittest

from ..support import BaseWebTest


@contextmanager
def notif_broken(app, event_cls):
    old = app.registry.notify

    def buggy(event):
        if not isinstance(event, event_cls):
            return old(event)
        raise Exception("boom")

    app.registry.notify = buggy
    yield
    app.registry.notify = old


class BaseEventTest(BaseWebTest):
    subscribed = tuple()
    events = []

    def setUp(self):
        super().setUp()
        del self.events[:]
        self.body = {"data": {"name": "de Paris"}}

    def tearDown(self):
        del self.events[:]
        super().tearDown()

    @classmethod
    def listener(cls, event):
        cls.events.append(event)

    @classmethod
    def make_app(cls, settings=None, config=None):
        settings = cls.get_app_settings(settings)
        config = Configurator(settings=settings)
        for event_cls in cls.subscribed:
            config.add_subscriber(cls.listener, event_cls)
        config.commit()
        return super().make_app(settings=settings, config=config)


class ResourceReadTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceRead,)

    def test_get_sends_read_event(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        object_id = resp.json["data"]["id"]
        object_url = self.get_item_url(object_id)
        self.app.get(object_url, headers=self.headers)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.READ.value)
        self.assertEqual(len(self.events[0].read_objects), 1)

    def test_plural_get_sends_read_event(self):
        self.app.get(self.plural_url, headers=self.headers)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.READ.value)
        self.assertEqual(len(self.events[0].read_objects), 0)

    def test_post_sends_read_if_id_already_exists(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        obj = resp.json["data"]
        body = {**self.body}
        body["data"]["id"] = obj["id"]

        # a second post with the same object id
        self.app.post_json(self.plural_url, body, headers=self.headers, status=200)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.READ.value)


class ResourceChangedTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceChanged,)

    def test_events_have_custom_representation(self):
        self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        self.assertEqual(
            repr(self.events[0]),
            "<ResourceChanged action=create " "uri={}>".format(self.plural_url),
        )

    def test_post_sends_create_action(self):
        self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)

    def test_put_sends_create_action(self):
        body = {**self.body}
        body["data"]["id"] = object_id = str(uuid.uuid4())
        object_url = self.get_item_url(object_id)
        self.app.put_json(object_url, body, headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)

    def test_put_event_has_sensible_timestamp(self):
        body = {**self.body}
        body["data"]["id"] = object_id = str(uuid.uuid4())
        object_url = self.get_item_url(object_id)
        resp = self.app.put_json(object_url, body, headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(self.events[0].payload["timestamp"], resp.json["data"]["last_modified"])

    def test_not_triggered_on_failed_put(self):
        body = {**self.body}
        body["data"]["id"] = object_id = str(uuid.uuid4())
        object_url = self.get_item_url(object_id)
        self.app.put_json(object_url, body, headers=self.headers)
        headers = {**self.headers, "If-Match": '"12345"'}
        self.app.put_json(object_url, body, headers=headers, status=412)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)

    def test_patch_sends_update_action(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        obj = resp.json["data"]
        object_url = self.get_item_url(obj["id"])

        self.app.patch_json(object_url, self.body, headers=self.headers, status=200)
        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload["action"], ACTIONS.UPDATE.value)

    def test_put_sends_update_action_if_object_exists(self):
        body = {**self.body}
        body["data"]["id"] = object_id = str(uuid.uuid4())
        object_url = self.get_item_url(object_id)
        self.app.put_json(object_url, body, headers=self.headers, status=201)

        body["data"]["more"] = "stuff"
        self.app.put_json(object_url, body, headers=self.headers, status=200)

        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload["action"], ACTIONS.UPDATE.value)

    def test_delete_sends_delete_action(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        obj = resp.json["data"]
        object_url = self.get_item_url(obj["id"])

        self.app.delete(object_url, headers=self.headers, status=200)
        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload["action"], ACTIONS.DELETE.value)

    def test_plural_delete_sends_delete_action(self):
        self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)

        self.app.delete(self.plural_url, headers=self.headers, status=200)

        self.assertEqual(len(self.events), 3)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(self.events[2].payload["action"], ACTIONS.DELETE.value)

    def test_request_fails_if_notify_fails(self):
        with notif_broken(self.app.app, ResourceChanged):
            self.app.post_json(self.plural_url, self.body, headers=self.headers, status=500)
        self.assertEqual(len(self.events), 0)

    def test_triggered_on_protected_resource(self):
        app = self.make_app(settings={"psilo_write_principals": "system.Authenticated"})
        app.post_json("/psilos", self.body, headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload["action"], ACTIONS.CREATE.value)

    def test_permissions_are_stripped_from_event_on_protected_resource(self):
        app = self.make_app(settings={"psilo_write_principals": "system.Authenticated"})
        resp = app.post_json("/psilos", self.body, headers=self.headers, status=201)
        obj = resp.json["data"]
        object_url = "/psilos/{}".format(obj["id"])
        app.patch_json(object_url, {"data": {"name": "De barcelona"}}, headers=self.headers)
        impacted_objects = self.events[-1].impacted_objects
        self.assertNotIn("__permissions__", impacted_objects[0]["new"])
        self.assertNotIn("__permissions__", impacted_objects[0]["old"])


class AfterResourceChangedTest(BaseEventTest, unittest.TestCase):
    subscribed = (AfterResourceChanged,)

    def test_request_succeeds_if_notify_fails(self):
        with notif_broken(self.app.app, AfterResourceChanged):
            self.app.post_json(self.plural_url, self.body, headers=self.headers)

        self.assertEqual(len(self.events), 0)


class AfterResourceReadTest(BaseEventTest, unittest.TestCase):
    subscribed = (AfterResourceRead,)

    def test_request_succeeds_if_notify_fails(self):
        with notif_broken(self.app.app, AfterResourceChanged):
            self.app.post_json(self.plural_url, self.body, headers=self.headers)

        self.assertEqual(len(self.events), 0)


class ImpactedObjectsTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceChanged,)

    def test_create_has_new_object_and_no_old_in_payload(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers)
        obj = resp.json["data"]
        impacted_objects = self.events[-1].impacted_objects
        self.assertEqual(len(impacted_objects), 1)
        self.assertNotIn("old", impacted_objects[0])
        self.assertEqual(impacted_objects[0]["new"], obj)

    def test_plural_delete_has_old_and_new_in_payload(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers)
        object1 = resp.json["data"]
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers)
        object2 = resp.json["data"]

        self.app.delete(self.plural_url, headers=self.headers, status=200)

        impacted_objects = self.events[-1].impacted_objects
        self.assertEqual(len(impacted_objects), 2)
        impacted_objects = sorted(impacted_objects, key=lambda x: x["old"]["last_modified"])
        self.assertEqual(impacted_objects[0]["old"], object1)
        self.assertEqual(impacted_objects[1]["old"], object2)
        self.assertEqual(impacted_objects[0]["new"]["deleted"], True)
        self.assertEqual(impacted_objects[1]["new"]["deleted"], True)
        self.assertEqual(impacted_objects[0]["old"]["id"], impacted_objects[0]["new"]["id"])
        self.assertEqual(impacted_objects[1]["old"]["id"], impacted_objects[1]["new"]["id"])
        deleted_ids = {impacted_objects[0]["new"]["id"], impacted_objects[1]["new"]["id"]}
        self.assertEqual(deleted_ids, {object1["id"], object2["id"]})

    def test_update_has_old_and_new_object(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        obj = resp.json["data"]
        object_url = self.get_item_url(obj["id"])
        self.app.patch_json(object_url, {"data": {"name": "en boite"}}, headers=self.headers)
        impacted_objects = self.events[-1].impacted_objects
        self.assertEqual(len(impacted_objects), 1)
        self.assertEqual(impacted_objects[0]["new"]["id"], obj["id"])
        self.assertEqual(impacted_objects[0]["new"]["id"], impacted_objects[0]["old"]["id"])
        self.assertEqual(impacted_objects[0]["old"]["name"], "de Paris")
        self.assertEqual(impacted_objects[0]["new"]["name"], "en boite")

    def test_delete_has_old_and_new_object_in_payload(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        obj = resp.json["data"]
        object_url = self.get_item_url(obj["id"])
        self.app.delete(object_url, headers=self.headers, status=200)

        impacted_objects = self.events[-1].impacted_objects
        self.assertEqual(len(impacted_objects), 1)
        self.assertEqual(impacted_objects[0]["old"], obj)
        self.assertEqual(impacted_objects[0]["new"]["id"], obj["id"])
        self.assertEqual(impacted_objects[0]["new"]["deleted"], True)


class BatchEventsTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceChanged, ResourceRead)

    def test_impacted_objects_are_merged(self):
        object_id = str(uuid.uuid4())
        object_url = self.get_item_url(object_id)
        body = {
            "defaults": {"method": "PUT", "path": object_url},
            "requests": [
                {"body": {"data": {"name": "foo"}}},
                {"body": {"data": {"name": "bar"}}},
                {"body": {"data": {"name": "baz"}}},
                {"method": "DELETE"},
            ],
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 3)

        create_event = self.events[0]
        self.assertEqual(create_event.payload["action"], "create")
        self.assertEqual(len(create_event.impacted_objects), 1)
        self.assertNotIn("old", create_event.impacted_objects[0])
        update_event = self.events[1]
        self.assertEqual(update_event.payload["action"], "update")
        impacted = update_event.impacted_objects
        self.assertEqual(len(impacted), 2)
        self.assertEqual(impacted[0]["old"]["name"], "foo")
        self.assertEqual(impacted[0]["new"]["name"], "bar")
        self.assertEqual(impacted[1]["old"]["name"], "bar")
        self.assertEqual(impacted[1]["new"]["name"], "baz")
        delete_event = self.events[2]
        self.assertEqual(delete_event.payload["action"], "delete")
        self.assertEqual(len(delete_event.impacted_objects), 1)
        self.assertIn("old", delete_event.impacted_objects[0])
        self.assertIn("new", delete_event.impacted_objects[0])

    def test_one_event_is_sent_per_resource(self):
        body = {
            "defaults": {"method": "POST", "body": self.body},
            "requests": [{"path": "/mushrooms"}, {"path": "/mushrooms"}, {"path": "/psilos"}],
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 2)

    def test_one_event_is_sent_per_parent_id(self):
        # /mushrooms is a Resource (see testapp.views) whose parent id depends
        # on the authenticated user.
        body = {
            "defaults": {"path": "/mushrooms", "method": "POST", "body": self.body},
            "requests": [
                {"headers": {"Authorization": "Basic bWF0OjE="}},
                {"headers": {"Authorization": "Basic dG90bzp0dXR1"}},
                {"headers": {"Authorization": "Basic bWF0OjE="}},
            ],
        }
        self.app.post_json("/batch", body, headers=self.headers)
        # Two different auth headers, thus two different parent_id:
        self.assertEqual(len(self.events), 2)

    def test_one_event_is_sent_per_action(self):
        body = {
            "defaults": {"path": "/mushrooms"},
            "requests": [
                {"method": "POST", "body": self.body},
                {"method": "DELETE"},
                {"method": "GET"},
            ],
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 3)

    def test_events_are_not_sent_if_subrequest_fails(self):
        patch = mock.patch.object(self.storage, "delete_all", side_effect=BackendError("boom"))
        patch.start()
        self.addCleanup(patch.stop)
        request_create = {"method": "POST", "body": self.body}
        request_delete_all = {"method": "DELETE", "body": self.body}
        body = {
            "defaults": {"path": self.plural_url},
            "requests": [request_create, request_delete_all],
        }
        self.app.post_json("/batch", body, headers=self.headers, status=503)
        self.assertEqual(len(self.events), 0)


class CascadingEventsTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceChanged, AfterResourceChanged)

    @classmethod
    def listener(cls, event):
        """Have the first "de Paris" event trigger a follow-up "de New York" event."""
        cls.events.append(event)
        if not isinstance(event, ResourceChanged):
            return
        if len(event.impacted_objects) > 1:
            raise ValueError("Too many events {}".format(event.impacted_objects))
        # An event without objects is impossible.
        assert len(event.impacted_objects) == 1
        if event.impacted_objects[0]["new"]["name"] == "de Paris":
            new_object = {"name": "de New York"}
            # Trying to match the Mushroom parent ID to test event grouping
            parent_id = event.request.prefixed_userid
            notify_resource_event(
                event.request, parent_id, event.payload["timestamp"], new_object, ACTIONS.CREATE
            )

    def test_event_can_trigger_other_event(self):
        self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        resource_changed_events = [e for e in self.events if isinstance(e, ResourceChanged)]
        self.assertEqual(len(resource_changed_events), 2)
        self.assertEqual(resource_changed_events[0].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(len(resource_changed_events[0].impacted_objects), 1)
        self.assertEqual(resource_changed_events[0].impacted_objects[0]["new"]["name"], "de Paris")
        self.assertEqual(resource_changed_events[1].payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(len(resource_changed_events[1].impacted_objects), 1)
        self.assertEqual(
            resource_changed_events[1].impacted_objects[0]["new"]["name"], "de New York"
        )

    def test_cascading_events_are_merged_in_after_resourced_changed(self):
        self.app.post_json(self.plural_url, self.body, headers=self.headers, status=201)
        relevant_events = [e for e in self.events if isinstance(e, AfterResourceChanged)]
        self.assertEqual(len(relevant_events), 1)
        merged_event = relevant_events[0]
        self.assertEqual(merged_event.payload["action"], ACTIONS.CREATE.value)
        self.assertEqual(len(merged_event.impacted_objects), 2)
        self.assertEqual(merged_event.impacted_objects[0]["new"]["name"], "de Paris")
        self.assertEqual(merged_event.impacted_objects[1]["new"]["name"], "de New York")


class BatchCascadeEventsTest(BaseEventTest, unittest.TestCase):
    subscribed = (ResourceChanged,)

    @classmethod
    def listener(cls, event):
        cls.events.append(event)

        # In these resource tests, the resource parent id is the user id.
        parent_id = event.request.prefixed_userid

        # Let's create a cascade of event. When a CREATE event is received,
        # we send an UPDATE one.
        if event.payload["action"] == "create":
            notify_resource_event(
                event.request, parent_id, event.payload["timestamp"], {"foo": "42"}, ACTIONS.UPDATE
            )

    def test_cascading_events_are_not_merged(self):
        resp = self.app.post_json(self.plural_url, self.body, headers=self.headers)
        existing = resp.json["data"]
        before = len(self.events)

        # In this batch request, the first request will lead to a CREATE event.
        # The above listener will trigger an UPDATE event, and we then make sure
        # that it wasn't be merged with the other UPDATED event related to the
        # second request (PATCH).
        self.app.post_json(
            "/batch",
            {
                "requests": [
                    {
                        "method": "POST",
                        "path": "/mushrooms",
                        "body": {"data": {"name": "b"}},
                    },
                    {
                        "method": "PATCH",
                        "path": "/mushrooms/" + existing["id"],
                        "body": {"data": {"foo": "bar"}},
                    },
                ],
            },
            headers=self.headers,
        )

        batch_events = self.events[before:]
        self.assertEqual(len(batch_events), 3)
        # Creation (first request)
        self.assertEqual(len(batch_events[0].impacted_objects), 1)
        self.assertEqual(batch_events[0].impacted_objects[0]["new"]["name"], "b")
        # Update (second request)
        self.assertEqual(len(batch_events[1].impacted_objects), 1)
        self.assertEqual(batch_events[1].impacted_objects[0]["new"]["foo"], "bar")
        # Update (cascade event)
        self.assertEqual(len(batch_events[2].impacted_objects), 1)
        self.assertEqual(batch_events[2].impacted_objects[0]["new"]["foo"], "42")


class DeprecatedAttributes(unittest.TestCase):
    def setUp(self):
        patch = mock.patch("warnings.warn")
        self.mocked_warnings = patch.start()
        self.addCleanup(patch.stop)

    def test_read_records_is_deprecated(self):
        event = ResourceRead(None, [], None)
        event.read_records

        message = "`read_records` is deprecated, use `read_objects` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)

    def test_impacted_records_is_deprecated(self):
        event = ResourceChanged(None, [], None)
        event.impacted_records

        message = "`impacted_records` is deprecated, use `impacted_objects` instead."
        self.mocked_warnings.assert_called_with(message, DeprecationWarning)


def load_from_config(config, prefix):
    class ClassListener:
        def __call__(self, event):
            pass

    return ClassListener()


@unittest.skipIf(not statsd.statsd_module, "statsd is not installed.")
class StatsDTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, *args, **kwargs):
        settings = super().get_app_settings(*args, **kwargs)
        if not statsd.statsd_module:
            return settings

        settings["statsd_url"] = "udp://localhost:8125"
        this_module = "tests.core.resource.test_events"
        settings["event_listeners"] = "test"
        settings["event_listeners.test.use"] = this_module
        return settings

    def test_statds_tracks_listeners_execution_duration(self):
        statsd_client = self.app.app.registry.statsd._client
        with mock.patch.object(statsd_client, "timing") as mocked:
            self.app.post_json(self.plural_url, {"data": {"name": "pouet"}}, headers=self.headers)
            timers = set(c[0][0] for c in mocked.call_args_list)
            self.assertIn("listeners.test", timers)



import unittest

from bravado.requests_client import RequestsClient
from bravado_core.request import IncomingRequest, unmarshal_request
from bravado_core.resource import build_resources
from bravado_core.response import OutgoingResponse, validate_response
from bravado_core.spec import Spec

from kinto.core.utils import json

from ..support import (
    MINIMALIST_BUCKET,
    MINIMALIST_COLLECTION,
    MINIMALIST_GROUP,
    MINIMALIST_RECORD,
    BaseWebTest,
)


class OpenAPITest(BaseWebTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.spec_dict = cls.app.get("/__api__").json
        bravado_config = {
            # Use models (Python classes) instead of dicts for #/definitions/{models}
            # use_models causes us to break in bravado-core 4.13.0,
            # probably because of
            # https://github.com/Yelp/bravado-core/pull/254, and we
            # don't actually use the generated models in our tests
            # here anyhow.
            "use_models": False
        }
        cls.spec = Spec.from_dict(
            cls.spec_dict, http_client=RequestsClient(), config=bravado_config
        )
        cls.resources = build_resources(cls.spec)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = ["kinto.plugins.history", "kinto.plugins.admin"]
        return settings

    def setUp(self):
        super().setUp()

        self.bucket = self.app.put_json(
            "/buckets/b1", MINIMALIST_BUCKET, headers=self.headers
        ).json

        self.group = self.app.put_json(
            "/buckets/b1/groups/g1", MINIMALIST_GROUP, headers=self.headers
        ).json

        self.collection = self.app.put_json(
            "/buckets/b1/collections/c1", MINIMALIST_COLLECTION, headers=self.headers
        ).json

        self.record = self.app.put_json(
            "/buckets/b1/collections/c1/records/r1", MINIMALIST_RECORD, headers=self.headers
        ).json

        # Create raw Bravado request
        self.request = IncomingRequest()
        self.request.url = ""
        self.request.data = ""
        self.request.path = {}
        self.request.headers = {}
        self.request.query = {}
        self.request._json = {}
        self.request.json = lambda: self.request._json

    def cast_bravado_response(self, response):
        resp = OutgoingResponse()
        resp.text = response.body

        # FIXME: Empty response fails with Bravado and Python 3
        # See https://github.com/Yelp/bravado-core/issues/134
        if not resp.text:
            resp.text = None

        resp.headers = response.headers
        # Response headers integer fields are not casted by default
        for k, v in resp.headers.items():
            try:
                resp.headers[k] = int(v)
            except ValueError:
                pass

        # FIXME: Drop charset from application/json response on Pyramid
        # See https://github.com/Pylons/pyramid/issues/2860
        resp.content_type = response.headers.get("Content-Type", "").split(";")[0]
        resp.json = lambda: response.json

        return resp

    def validate_request_call(self, op, **kargs):
        params = unmarshal_request(self.request, op)
        response = self.app.request(
            op.path_name.format_map(params),
            body=json.dumps(self.request.json()).encode(),
            method=op.http_method.upper(),
            headers=self.headers,
            **kargs,
        )
        schema = self.spec.deref(op.op_spec["responses"][str(response.status_code)])
        casted_resp = self.cast_bravado_response(response)
        validate_response(schema, op, casted_resp)
        return response



import builtins
import logging
import os
import sys
import tempfile
import unittest
from io import StringIO
from unittest import mock

import pytest

from kinto import __version__ as kinto_version
from kinto.__main__ import DEFAULT_LOG_FORMAT, main


fd, TEMP_KINTO_INI = tempfile.mkstemp(prefix="kinto_config", suffix=".ini")
# The above call to mkstemp returns a handle to an open file. On Windows,
# attempting to call os.remove on this file will result in an error unless it is
# manually closed.
os.fdopen(fd).close()


class TestMain(unittest.TestCase):
    def setUp(self):
        try:
            os.remove(TEMP_KINTO_INI)
        except FileNotFoundError:
            pass

    def test_cli_init_generates_configuration(self):
        res = main(
            ["init", "--ini", TEMP_KINTO_INI, "--backend", "memory", "--cache-backend", "memory"]
        )
        assert res == 0
        assert os.path.exists(TEMP_KINTO_INI)

    def test_cli_init_returns_if_file_exists(self):
        with open(TEMP_KINTO_INI, "w") as f:
            f.write("exists")
        with mock.patch("sys.stderr", new_callable=StringIO) as mock_stderr:
            res = main(["init", "--ini", TEMP_KINTO_INI, "--backend", "memory"])
            assert res == 1
            content = mock_stderr.getvalue()
            assert "{} already exists.".format(TEMP_KINTO_INI) in content

    def test_cli_init_asks_for_backend_if_not_specified(self):
        with mock.patch("kinto.__main__.input", create=True, return_value="1"):
            res = main(["init", "--ini", TEMP_KINTO_INI])
            assert res == 0
        with open(TEMP_KINTO_INI) as f:
            content = f.read()
        assert "postgresql" in content

    def test_cli_init_asks_until_backend_is_valid(self):
        with mock.patch("kinto.__main__.input", create=True, side_effect=["10", "1", "1"]):
            res = main(["init", "--ini", TEMP_KINTO_INI])
            assert res == 0
            with open(TEMP_KINTO_INI) as f:
                content = f.read()
            assert "postgresql" in content

    def test_cli_init_asks_until_cache_backend_is_valid(self):
        with mock.patch("kinto.__main__.input", create=True, side_effect=["1", "10", "2"]):
            res = main(["init", "--ini", TEMP_KINTO_INI])
            assert res == 0
            with open(TEMP_KINTO_INI) as f:
                content = f.read()
            assert "memcached" in content

    def test_fails_if_not_enough_args(self):
        with mock.patch("sys.stderr", new_callable=StringIO) as mock_stderr:
            with pytest.raises(SystemExit) as excinfo:
                main([])
            assert "arguments are required: subcommand" in mock_stderr.getvalue()
            assert excinfo.value.code == 2

    def test_cli_init_installs_postgresql_dependencies_if_needed(self):
        realimport = builtins.__import__

        def psycopg2_missing(name, *args, **kwargs):
            if name == "psycopg2":
                raise ImportError()
            else:
                return realimport(name, *args, **kwargs)

        with mock.patch("builtins.__import__", side_effect=psycopg2_missing):
            with mock.patch(
                "kinto.__main__.subprocess.check_call", return_value=None
            ) as mocked_pip:
                with mock.patch("kinto.__main__.input", create=True, return_value="1"):
                    res = main(["init", "--ini", TEMP_KINTO_INI])
                    assert res == 0
                    assert mocked_pip.call_count == 1

    def test_cli_init_installs_memcached_dependencies_if_needed(self):
        realimport = builtins.__import__

        def memcached_missing(name, *args, **kwargs):
            if name == "memcache":
                raise ImportError()
            else:
                return realimport(name, *args, **kwargs)

        with mock.patch("builtins.__import__", side_effect=memcached_missing):
            with mock.patch(
                "kinto.__main__.subprocess.check_call", return_value=None
            ) as mocked_pip:
                with mock.patch("kinto.__main__.input", create=True, return_value="2"):
                    res = main(["init", "--ini", TEMP_KINTO_INI, "--backend", "memory"])
                    assert res == 0
                    assert mocked_pip.call_count == 1

    def test_main_takes_sys_argv_by_default(self):
        testargs = [
            "prog",
            "init",
            "--ini",
            TEMP_KINTO_INI,
            "--backend",
            "memory",
            "--cache-backend",
            "memory",
        ]
        with mock.patch.object(sys, "argv", testargs):
            main()

        with open(TEMP_KINTO_INI) as f:
            content = f.read()
        assert "memory" in content

    def test_cli_migrate_command_runs_init_schema(self):
        with mock.patch("kinto.__main__.core_scripts.migrate") as mocked_migrate:
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(["migrate", "--ini", TEMP_KINTO_INI])
            assert res == 0
            assert mocked_migrate.call_count == 1

    def test_cli_rebuild_quotas_run_rebuild_quotas_script(self):
        with mock.patch("kinto.__main__.kinto_scripts.rebuild_quotas") as reb_quo:
            reb_quo.return_value = mock.sentinel.reb_quo_code
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(["rebuild-quotas", "--ini", TEMP_KINTO_INI])
            assert res == mock.sentinel.reb_quo_code
            assert reb_quo.call_count == 1

    def test_cli_start_runs_pserve(self):
        with mock.patch("kinto.__main__.pserve.main") as mocked_pserve:
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(["start", "--ini", TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1

    def test_cli_start_with_quiet_option_runs_pserve_with_quiet(self):
        with mock.patch("kinto.__main__.pserve.main") as mocked_pserve:
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(["start", "-q", "--ini", TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1
            assert mocked_pserve.call_args[1]["argv"][1] == "-q"

    def test_cli_start_with_verbose_option_runs_pserve_with_verbose(self):
        with mock.patch("kinto.__main__.pserve.main") as mocked_pserve:
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(["start", "-v", "--ini", TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1
            assert mocked_pserve.call_args[1]["argv"][1] == "-v"

    def test_cli_start_with_reload_runs_pserve_with_reload(self):
        with mock.patch("kinto.__main__.pserve.main") as mocked_pserve:
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(["start", "--reload", "--ini", TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1
            assert "--reload" in mocked_pserve.call_args[1]["argv"]

    def test_cli_create_user_runs_account_script(self):
        with mock.patch(
            "kinto.__main__.accounts_scripts.create_user", return_value=0
        ) as mocked_create_user:
            res = main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            assert res == 0
            res = main(
                ["create-user", "--ini", TEMP_KINTO_INI, "-u", "username", "-p", "password"]
            )
            assert res == 0
            mocked_create_user.call_count == 1

    def test_cli_can_display_kinto_version(self):
        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            res = main(["version"])
            assert mock_stdout.getvalue() == "{}\n".format(kinto_version)
            assert res == 0

    def test_cli_can_configure_logger_in_quiet(self):
        with mock.patch("kinto.__main__.logging") as mocked_logging:
            main(
                [
                    "init",
                    "-q",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            mocked_logging.basicConfig.assert_called_with(
                level=mocked_logging.CRITICAL, format=DEFAULT_LOG_FORMAT
            )

    def test_cli_can_configure_logger_in_debug(self):
        with mock.patch("kinto.__main__.logging") as mocked_logging:
            main(
                [
                    "init",
                    "--debug",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            mocked_logging.basicConfig.assert_called_with(
                level=mocked_logging.DEBUG, format=DEFAULT_LOG_FORMAT
            )

    def test_cli_use_default_logging_logger(self):
        with mock.patch("kinto.__main__.logging") as mocked_logging:
            main(
                [
                    "init",
                    "--ini",
                    TEMP_KINTO_INI,
                    "--backend",
                    "memory",
                    "--cache-backend",
                    "memory",
                ]
            )
            mocked_logging.basicConfig.assert_called_with(
                level=logging.INFO, format=DEFAULT_LOG_FORMAT
            )

    def test_cli_flush_cache_command_runs_flush_cache_script(self):
        # Build a temporary ini file.
        res = main(
            ["init", "--ini", TEMP_KINTO_INI, "--backend", "memory", "--cache-backend", "memory"]
        )
        assert res == 0
        with mock.patch("kinto.__main__.core_scripts.flush_cache") as mocked_cache_script:
            res = main(["flush-cache", "--ini", TEMP_KINTO_INI])
            assert res == 0
            assert mocked_cache_script.call_count == 1



from cornice import Service
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.events import ServerFlushed


flush = Service(name="flush", description="Clear database content", path="/__flush__")


@flush.post(permission=NO_PERMISSION_REQUIRED)
def flush_post(request):
    request.registry.storage.flush()
    request.registry.permission.flush()
    request.registry.cache.flush()
    event = ServerFlushed(request)
    request.registry.notify(event)

    request.response.status = 202
    return {}


def includeme(config):
    config.add_api_capability(
        "flush_endpoint",
        description="The __flush__ endpoint can be used to remove " "all data from all backends.",
        url="https://kinto.readthedocs.io/en/latest/api/1.x/flush.html",
    )
    config.add_cornice_service(flush)



from kinto.core.testing import unittest
from kinto.schema_validation import validate_schema

from .support import BaseWebTest


BUCKET_URL = "/buckets/blog"
COLLECTION_URL = "/buckets/blog/collections/articles"
RECORDS_URL = "/buckets/blog/collections/articles/records"


SCHEMA = {
    "title": "Blog post schema",
    "type": "object",
    "properties": {
        "title": {"type": "string"},
        "body": {"type": "string"},
        "file": {
            "type": "object",
            "properties": {"size": {"type": "number"}, "name": {"type": "string"}},
        },
        "tags": {"type": "array", "items": {"type": "string"}},
    },
    "required": ["title"],
}

VALID_RECORD = {"title": "About us", "body": "<h1>About</h1>"}


class DeactivatedSchemaTest(BaseWebTest, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, "type": "Washmachine"}
        self.app.put(BUCKET_URL, headers=self.headers)
        self.app.put(COLLECTION_URL, headers=self.headers)
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": newschema}}, headers=self.headers, status=400
        )
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json["message"])

    def test_records_are_not_invalid_if_do_not_match_schema(self):
        self.app.put_json(BUCKET_URL, headers=self.headers)
        self.app.put(COLLECTION_URL, headers=self.headers)
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": SCHEMA}}, headers=self.headers
        )
        self.collection = resp.json["data"]

        resp = self.app.post_json(
            RECORDS_URL,
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
            status=201,
        )
        self.assertNotIn("schema", resp.json["data"])


class BaseWebTestWithSchema(BaseWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_collection_schema_validation"] = "True"
        return settings

    def setUp(self):
        super().setUp()
        self.app.put(BUCKET_URL, headers=self.headers)
        self.app.put(COLLECTION_URL, headers=self.headers)


class MissingSchemaTest(BaseWebTestWithSchema, unittest.TestCase):
    def test_attribute_is_none_if_no_schema_defined(self):
        resp = self.app.get(COLLECTION_URL, headers=self.headers)
        self.assertIsNone(resp.json["data"].get(""))

    def test_accepts_any_kind_of_record(self):
        record = {"title": "Troll"}
        self.app.post_json(RECORDS_URL, {"data": record}, headers=self.headers, status=201)
        record = {"author": {"age": 32, "status": "captain"}}
        self.app.post_json(RECORDS_URL, {"data": record}, headers=self.headers, status=201)


class InvalidSchemaTest(BaseWebTestWithSchema, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, "type": "Washmachine"}
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": newschema}}, headers=self.headers, status=400
        )
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json["message"])

    def test_extra_unknown_required_property(self):
        schema = {**SCHEMA, "required": ["unknown"]}
        self.app.put_json(COLLECTION_URL, {"data": {"schema": schema}}, headers=self.headers)

        record = {"title": "bug 1243"}
        r = self.app.post_json(RECORDS_URL, {"data": record}, headers=self.headers, status=400)
        self.assertEqual("'unknown' is a required property", r.json["message"])

        # With bug Kinto/kinto#1243, the second call would crash.
        self.app.post_json(RECORDS_URL, {"data": record}, headers=self.headers, status=400)


class RecordsValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": SCHEMA}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_validate_schema(self):
        validate_schema(VALID_RECORD, SCHEMA, id_field="id")

    def test_empty_record_can_be_validated(self):
        self.app.post_json(RECORDS_URL, {"data": {}}, headers=self.headers, status=400)

    def test_records_are_valid_if_match_schema(self):
        self.app.post_json(RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers, status=201)

    def test_records_are_invalid_if_do_not_match_schema(self):
        self.app.post_json(
            RECORDS_URL,
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
            status=400,
        )

    def test_records_are_invalid_if_do_not_match_schema_in_array(self):
        self.app.post_json(
            RECORDS_URL, {"data": {**VALID_RECORD, "tags": [0]}}, headers=self.headers, status=400
        )

    def test_records_are_validated_on_patch(self):
        resp = self.app.post_json(
            RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers, status=201
        )
        record_id = resp.json["data"]["id"]
        self.app.patch_json(
            "{}/{}".format(RECORDS_URL, record_id),
            {"data": {"title": 3.14}},
            headers=self.headers,
            status=400,
        )

    def test_records_are_validated_on_put(self):
        resp = self.app.post_json(
            RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers, status=201
        )
        record_id = resp.json["data"]["id"]
        self.app.put_json(
            "{}/{}".format(RECORDS_URL, record_id),
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
            status=400,
        )

    def test_validation_error_response_provides_details(self):
        resp = self.app.post_json(
            RECORDS_URL,
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
            status=400,
        )
        self.assertIn("'title' is a required property", resp.json["message"])
        self.assertEqual(resp.json["details"][0]["name"], "title")

    def test_validation_error_response_provides_details_and_path(self):
        resp = self.app.post_json(
            RECORDS_URL,
            {"data": {"title": "Some title", "file": {"size": "hi!"}}},
            headers=self.headers,
            status=400,
        )
        self.assertIn("size in body: 'hi!' is not of type 'number'", resp.json["message"])
        self.assertEqual(resp.json["details"][0]["name"], "size")

    def test_records_of_other_bucket_are_not_impacted(self):
        self.app.put_json("/buckets/cms", headers=self.headers)
        self.app.put_json("/buckets/cms/collections/articles", headers=self.headers)
        self.app.post_json(
            "/buckets/cms/collections/articles/records",
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
        )

    def test_records_receive_the_schema_as_attribute(self):
        resp = self.app.post_json(
            RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers, status=201
        )
        self.assertEqual(resp.json["data"]["schema"], self.collection["last_modified"])

    def test_records_can_filtered_by_schema_version(self):
        self.app.post_json(RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers)
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": SCHEMA}}, headers=self.headers
        )
        schema_version = resp.json["data"]["last_modified"]
        self.app.post_json(RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers)

        resp = self.app.get(
            RECORDS_URL + "?min_schema={}".format(schema_version), headers=self.headers
        )
        self.assertEqual(len(resp.json["data"]), 1)


class ExtraPropertiesValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        schema = {**SCHEMA, "additionalProperties": False}
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": schema}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_record_can_be_validated_on_post(self):
        self.app.post_json(RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers)

    def test_record_can_be_validated_on_put(self):
        record_id = "5443d83f-852a-481a-8e9d-5aa804b05b08"
        self.app.put_json(
            "{}/{}".format(RECORDS_URL, record_id), {"data": VALID_RECORD}, headers=self.headers
        )

    def test_records_are_validated_on_patch(self):
        record_id = "5443d83f-852a-481a-8e9d-5aa804b05b08"
        record_url = "{}/{}".format(RECORDS_URL, record_id)
        resp = self.app.put_json(record_url, {"data": VALID_RECORD}, headers=self.headers)
        record = resp.json["data"]
        assert "schema" in record
        record["title"] = "hey"
        self.app.patch_json(record_url, {"data": record}, headers=self.headers)

    def test_additional_properties_are_rejected(self):
        record_id = "5443d83f-852a-481a-8e9d-5aa804b05b08"
        record = {**VALID_RECORD, "extra": "blah!"}
        resp = self.app.put_json(
            "{}/{}".format(RECORDS_URL, record_id),
            {"data": record},
            headers=self.headers,
            status=400,
        )
        assert "'extra' was unexpected)" in resp.json["message"]


class InternalRequiredProperties(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        # See bug Kinto/kinto#1244
        schema = {**SCHEMA, "required": ["id", "schema", "last_modified"]}
        self.app.put_json(COLLECTION_URL, {"data": {"schema": schema}}, headers=self.headers)

    def test_record_can_be_validated_with_minimum_fields(self):
        self.app.post_json(RECORDS_URL, {"data": {}}, headers=self.headers)

    def test_record_can_be_validated_with_every_fields(self):
        self.app.post_json(
            RECORDS_URL,
            {
                "data": {
                    "id": "abc",
                    "last_modified": 1234,
                    "schema": 42,
                    "title": "b",
                    "name": "n",
                }
            },
            headers=self.headers,
        )

    def test_record_can_be_validated_with_id_and_last_modified(self):
        self.app.post_json(
            RECORDS_URL, {"data": {"id": "abc", "last_modified": 1234}}, headers=self.headers
        )

    def test_record_validation_can_reject_records(self):
        self.app.post_json(
            RECORDS_URL,
            {"data": {"id": "abc", "last_modified": 1234, "body": 2}},
            headers=self.headers,
            status=400,
        )


class ValidateIDField(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        # See bug Kinto/kinto#1942
        schema = {"type": "object", "properties": {"id": {"type": "string", "pattern": "^[0-7]$"}}}
        self.app.put_json(COLLECTION_URL, {"data": {"schema": schema}}, headers=self.headers)

    def test_record_id_is_accepted_if_valid(self):
        self.app.post_json(RECORDS_URL, {"data": {"id": "1"}}, headers=self.headers)

    def test_record_id_is_rejected_if_does_not_match(self):
        self.app.post_json(RECORDS_URL, {"data": {"id": "a"}}, headers=self.headers, status=400)


class BucketRecordSchema(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.app.put_json(BUCKET_URL, {"data": {"record:schema": SCHEMA}}, headers=self.headers)

    def test_records_are_valid_if_match_schema(self):
        self.app.post_json(RECORDS_URL, {"data": VALID_RECORD}, headers=self.headers, status=201)

    def test_records_are_validated_on_batch(self):
        resp = self.app.post_json(
            "/batch",
            {
                "defaults": {"path": RECORDS_URL, "method": "POST"},
                "requests": [
                    {"body": {"data": VALID_RECORD}},
                    {"body": {"data": {"body": "<h1>Without title</h1>"}}},
                ],
            },
            headers=self.headers,
        )
        assert resp.json["responses"][0]["status"] == 201
        assert resp.json["responses"][1]["status"] == 400

    def test_records_are_invalid_if_do_not_match_schema(self):
        self.app.post_json(
            RECORDS_URL,
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
            status=400,
        )


class BothBucketAndCollectionSchemas(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.app.put_json(COLLECTION_URL, {"data": {"schema": SCHEMA}}, headers=self.headers)
        other_schema = {"type": "object", "properties": {"filters": {"type": "string"}}}
        self.app.put_json(
            BUCKET_URL, {"data": {"record:schema": other_schema}}, headers=self.headers
        )

    def test_records_are_valid_if_match_both_schemas(self):
        self.app.post_json(
            RECORDS_URL,
            {"data": {"filters": "1 == 1", **VALID_RECORD}},
            headers=self.headers,
            status=201,
        )

    def test_records_are_invalid_if_do_not_match_collection_schema(self):
        self.app.post_json(
            RECORDS_URL,
            {"data": {"body": "<h1>Without title</h1>"}},
            headers=self.headers,
            status=400,
        )

    def test_records_are_invalid_if_do_not_match_bucket_schema(self):
        self.app.post_json(
            RECORDS_URL,
            {"data": {"filters": 42, **VALID_RECORD}},
            headers=self.headers,
            status=400,
        )


SCHEMA_UNRESOLVABLE = {"properties": {"title": {"$ref": "#/definitions/title"}}}


class RecordsUnresolvableTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(
            COLLECTION_URL, {"data": {"schema": SCHEMA_UNRESOLVABLE}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_unresolvable_errors_handled(self):
        self.app.post_json(RECORDS_URL, {"data": {"title": "b"}}, headers=self.headers, status=400)


class BucketUnresolvableRecordSchema(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.app.put_json(
            BUCKET_URL, {"data": {"record:schema": SCHEMA_UNRESOLVABLE}}, headers=self.headers
        )

    def test_records_are_valid_if_match_schema(self):
        self.app.post_json(RECORDS_URL, {"data": {"title": "b"}}, headers=self.headers, status=400)


class RecordsWithLargeNumbers(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.app.put_json(COLLECTION_URL, {"data": {"schema": SCHEMA}}, headers=self.headers)

    def test_record_with_number_less_than_64_bits(self):
        size = 2**63
        self.app.post_json(
            RECORDS_URL,
            {"data": {"title": "Very large file", "file": {"size": size}}},
            headers=self.headers,
            status=201,
        )

    def test_record_with_number_greater_than_64_bits(self):
        size = 2**65
        self.app.post_json(
            RECORDS_URL,
            {"data": {"title": "Very large file", "file": {"size": size}}},
            headers=self.headers,
            status=201,
        )



import logging

from pyramid.settings import asbool


logger = logging.getLogger(__name__)


__HEARTBEAT_KEY__ = "__heartbeat__"


class PermissionBase:
    def __init__(self, *args, **kwargs):
        pass

    def initialize_schema(self, dry_run=False):
        """Create every necessary objects (like tables or indices) in the
        backend.

        This is executed with the ``kinto migrate`` command.

        :param bool dry_run: simulate instead of executing the operations.
        """
        raise NotImplementedError

    def flush(self):
        """Delete all data stored in the permission backend."""
        raise NotImplementedError

    def add_user_principal(self, user_id, principal):
        """Add an additional principal to a user.

        :param str user_id: The user_id to add the principal to.
        :param str principal: The principal to add.
        """
        raise NotImplementedError

    def remove_user_principal(self, user_id, principal):
        """Remove an additional principal from a user.

        :param str user_id: The user_id to remove the principal to.
        :param str principal: The principal to remove.
        """
        raise NotImplementedError

    def remove_principal(self, principal):
        """Remove a principal from every user.

        :param str principal: The principal to remove.
        """
        raise NotImplementedError

    def get_user_principals(self, user_id):
        """Return the set of additionnal principals given to a user.

        :param str user_id: The user_id to get the list of groups for.
        :returns: The list of group principals the user is in.
        :rtype: set

        """
        raise NotImplementedError

    def add_principal_to_ace(self, object_id, permission, principal):
        """Add a principal to an Access Control Entry.

        :param str object_id: The object to add the permission principal to.
        :param str permission: The permission to add the principal to.
        :param str principal: The principal to add to the ACE.
        """
        raise NotImplementedError

    def remove_principal_from_ace(self, object_id, permission, principal):
        """Remove a principal to an Access Control Entry.

        :param str object_id: The object to remove the permission principal to.
        :param str permission: The permission that should be removed.
        :param str principal: The principal to remove to the ACE.
        """
        raise NotImplementedError

    def get_object_permission_principals(self, object_id, permission):
        """Return the set of principals of a bound permission
        (unbound permission + object id).

        :param str object_id: The object_id the permission is set to.
        :param str permission: The permission to query.
        :returns: The list of user principals
        :rtype: set

        """
        raise NotImplementedError

    def get_accessible_objects(self, principals, bound_permissions=None, with_children=True):
        """Return the list of objects where the specified `principals`
        have some permissions.

        If `bound_permissions` parameter is specified, the list is limited to
        the specified object or permissions.

        :param list principals: List of user principals
        :param list bound_permissions: An optional list of tuples
            (object_id, permission) to limit the results.
            The object ids can be a pattern, (e.g. ``*``, ``'/my/articles*'``).
        :param bool with_children: Include the children of object ids or not.

        :returns: A mapping whose keys are the object_ids and the values are
            the related list of permissions.
        :rtype: dict
        """
        raise NotImplementedError

    def get_authorized_principals(self, bound_permissions):
        """Return the full set of authorized principals for a list of bound
        permissions (object + permission).

        :param str object_id: The object_id the permission is set to.
        :param list bound_permissions: An list of tuples
            (object_id, permission) to be fetched.
        :returns: The list of user principals
        :rtype: set

        """
        raise NotImplementedError

    def check_permission(self, principals, bound_permissions):
        """Test if a principal set have got a permission on an object.

        :param set principals:
            A set of user principals to test the permission against.
        :param list bound_permissions: An list of tuples
            (object_id, permission) to be checked.
        :rtype: bool
        """
        principals = set(principals)
        authorized = self.get_authorized_principals(bound_permissions)
        return len(authorized & principals) > 0

    def get_object_permissions(self, object_id, permissions=None):
        return self.get_objects_permissions([object_id], permissions)[0]

    def get_objects_permissions(self, objects_ids, permissions=None):
        """Return a list of mapping, for each object id specified, with the
        set of principals for each permission.

        :param list objects_ids: The list of object_ids.
        :param list permissions: Optional list of permissions to limit the
            results. If not specified, retrieve all.
        :returns: A list of dictionnaries with the list of user principals for
            each object permission.
        :rtype: list
        """
        raise NotImplementedError

    def replace_object_permissions(self, object_id, permissions):
        """Update the set of principals allowed to perform some actions on an
        object.

        The only update of permissions allowed by Kinto's API is
        complete replacement of some permission (e.g. I don't know who
        was previously allowed to read this object, but now it's Joe,
        Frank, and Paul). This method implements that, completely
        replacing the set of principals who are granted the given
        permissions (while leaving other permissions alone).

        :param str object_id: The object to replace permissions on.
        :param permissions: A dict of perm -> principals (where
        principals is an iterable of individuals) to be granted on
        this object.

        """
        raise NotImplementedError

    def delete_object_permissions(self, *object_id_list):
        """Delete all listed object permissions.

        :param str object_id: Remove given objects permissions.
        """
        raise NotImplementedError


def heartbeat(backend):
    def ping(request):
        """Test the permission backend is operational.

        :param request: current request object
        :type request: :class:`~pyramid:pyramid.request.Request`
        :returns: ``True`` is everything is ok, ``False`` otherwise.
        :rtype: bool
        """
        try:
            if asbool(request.registry.settings.get("readonly")):
                # Do not try to write in readonly mode.
                backend.get_user_principals(__HEARTBEAT_KEY__)
            else:
                backend.add_user_principal(__HEARTBEAT_KEY__, "alive")
                backend.remove_user_principal(__HEARTBEAT_KEY__, "alive")
        except Exception:
            logger.exception("Heartbeat Error")
            return False
        return True

    return ping
